#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
from abc import abstractmethod, ABCMeta

from ai_flow.meta.artifact_meta import ArtifactMeta
from ai_flow.endpoint.server.high_availability import Member
from typing import Text, Union, List, Optional

from ai_flow.meta.metric_meta import MetricMeta, MetricSummary


class AbstractStore(object):
    __metaclass__ = ABCMeta

    def __init__(self):
        pass

    '''
        model api
    '''

    @abstractmethod
    def get_model_relation_by_id(self, model_id):
        """
        get a specific model relation in metadata store by model id.

        :param model_id: the model id
        :return: A single :py:class:`ai_flow.meta.model_relation_meta.ModelRelationMeta` object if the model relation
                 exists, Otherwise, returns None if the model relation does not exist.
        """
        pass

    @abstractmethod
    def get_model_relation_by_name(self, model_name):
        """
        get a specific model relation in metadata store by model name.

        :param model_name: the model name
        :return: A single :py:class:`ai_flow.meta.model_relation_meta.ModelRelationMeta` object if the model relation
                 exists, Otherwise, returns None if the model relation does not exist.
        """
        pass

    @abstractmethod
    def list_model_relation(self, page_size, offset):
        """
        List registered model relations in metadata store.

        :param page_size: the limitation of the listed model relations.
        :param offset: the offset of listed model relations.
        :return: List of :py:class:`ai_flow.meta.model_relation_meta.ModelRelationMeta` objects,
                 return None if no model relations to be listed.
        """
        pass

    @abstractmethod
    def register_model_relation(self, name: Text,
                                project_id: int):
        """
        register a model relation in metadata store

        :param name: the name of the model
        :param project_id: the project id which the model corresponded to.
        :return: A single :py:class:`ai_flow.meta.model_relation_meta.ModelRelationMeta` object.
        """
        pass

    def delete_model_relation_by_id(self, model_id):
        """
        Delete the registered model by model id .

        :param model_id: the model id
        :return: Status.OK if the model is successfully deleted, Status.ERROR if the model does not exist otherwise.
        """
        pass

    def delete_model_relation_by_name(self, model_name):
        """
        Delete the registered model by model name .

        :param model_name: the model name
        :return: Status.OK if the model is successfully deleted, Status.ERROR if the model does not exist otherwise.
        """
        pass

    '''
        model version api
    '''

    @abstractmethod
    def get_model_version_relation_by_version(self, version_name, model_id):
        """
        get a specific model version relation in metadata store by the model version name.

        :param version_name: the model version name
        :param model_id: the model id corresponded to the model version
        :return: A single :py:class:`ai_flow.meta.model_relation_meta.ModelVersionRelationMeta` object
                 if the model version exists, Otherwise, returns None if the model version does not exist.
        """
        pass

    @abstractmethod
    def register_model_version_relation(self, version, model_id, project_snapshot_id):
        """
        register a model version relation in metadata store.

        :param version: the specific model version
        :param model_id: the model id corresponded to the model version
        :param project_snapshot_id: the project snapshot id corresponded to the model version
        :return: A single :py:class:`ai_flow.meta.model_relation_meta.ModelVersionRelationMeta` object.
        """
        pass

    @abstractmethod
    def list_model_version_relation(self, model_id, page_size, offset):
        """
        List registered model version relations in metadata store.

        :param model_id: the model id corresponded to the model version
        :param page_size: the limitation of the listed model version relations.
        :param offset: the offset of listed model version relations.
        :return: List of :py:class:`ai_flow.meta.model_relation_meta.ModelRelationMeta` objects,
                 return None if no model version relations to be listed.
        """
        pass

    @abstractmethod
    def delete_model_version_relation_by_version(self, version, model_id):
        """
        Delete the registered model version by model version name .

        :param version: the model version name
        :param model_id: the model id corresponded to the model version
        :return: Status.OK if the model version is successfully deleted,
                 Status.ERROR if the model version does not exist otherwise.
        """
        pass

    '''
        dataset api
    '''

    @abstractmethod
    def get_dataset_by_id(self, dataset_id):
        """
        get a specific dataset in metadata store by dataset id.

        :param dataset_id: the dataset id
        :return: A single :py:class:`ai_flow.meta.dataset_meta.DatasetMeta` object if the dataset exists,
                 Otherwise, returns None if the dataset does not exist.
        """
        pass

    @abstractmethod
    def get_dataset_by_name(self, dataset_name):
        """
        get a specific dataset in metadata store by dataset name.

        :param dataset_name: the dataset name
        :return: A single :py:class:`ai_flow.meta.dataset_meta.DatasetMeta` object if the dataset exists,,
                 Otherwise, returns None if the dataset does not exist.
        """
        pass

    @abstractmethod
    def list_datasets(self, page_size, offset):
        """
        List registered datasets in metadata store.

        :param page_size: the limitation of the listed datasets.
        :param offset: the offset of listed datasets.
        :return: List of :py:class:`ai_flow.meta.dataset_meta.DatasetMeta` objects,
                 return None if no datasets to be listed.
        """
        pass

    @abstractmethod
    def register_dataset(self, name, data_format,
                         description, uri,
                         properties,
                         name_list, type_list):
        """
        register an dataset in metadata store.

        :param name: the name of the dataset
        :param data_format: the data_format of the dataset
        :param description: the description of the dataset
        :param uri: the uri of the dataset
        :param properties: the properties of the dataset
        :param name_list: the name list of dataset's schema
        :param type_list: the type list corresponded to the name list of dataset's schema
        :return: A single :py:class:`ai_flow.meta.dataset_meta.DatasetMeta` object.
        """
        pass

    def delete_dataset_by_id(self, dataset_id):
        """
        Delete the registered dataset by dataset id .

        :param dataset_id: the dataset id
        :return: Status.OK if the dataset is successfully deleted, Status.ERROR if the dataset does not exist otherwise.
        """
        pass

    def delete_dataset_by_name(self, dataset_name):
        """
        Delete the registered dataset by dataset name .

        :param dataset_name: the dataset name
        :return: Status.OK if the dataset is successfully deleted, Status.ERROR if the dataset does not exist otherwise.
        """
        pass

    '''
        project api
    '''

    @abstractmethod
    def get_project_by_id(self, project_id):
        """
        get a specific project in metadata store by project id

        :param project_id: the project id
        :return: A single :py:class:`ai_flow.meta.project.ProjectMeta` object if the project exists,
                 Otherwise, returns None if the project does not exist.
        """
        pass

    @abstractmethod
    def get_project_by_name(self, project_name):
        """
        get a specific project in metadata store by project name
        :param project_name: the project name
        :return: A single :py:class:`ai_flow.meta.project.ProjectMeta` object if the project exists,
                 Otherwise, returns None if the project does not exist.
        """
        pass

    @abstractmethod
    def register_project(self, name, uri, properties):
        """
        register a project in metadata store.

        :param name: the name of the project
        :param uri: the uri of the project
        :param properties: the properties of the project
        :return: A single :py:class:`ai_flow.meta.project.ProjectMeta` object.
        """
        pass

    @abstractmethod
    def list_project(self, page_size, offset):
        """
        List registered projects in metadata store.

        :param page_size: the limitation of the listed projects.
        :param offset: the offset of listed projects.
        :return: List of :py:class:`ai_flow.meta.project_meta.ProjectMeta` objects,
                 return None if no projects to be listed.
        """
        pass

    @abstractmethod
    def delete_project_by_id(self, project_id):
        """
        Delete the registered project by project id .

        :param project_id: the project id
        :return: Status.OK if the project is successfully deleted, Status.ERROR if the project does not exist otherwise.
        """
        pass

    def delete_project_by_name(self, project_name):
        """
        Delete the registered project by project name .

        :param project_name: the project name
        :return: Status.OK if the project is successfully deleted, Status.ERROR if the project does not exist otherwise.
        """
        pass

    '''
        workflow api
    '''

    def register_workflow(self, name, project_id, properties=None):
        """
        Register a workflow in metadata store.

        :param name: the workflow name
        :param project_id: the id of project which contains the workflow
        :param properties: the workflow properties
        """
        pass

    def get_workflow_by_name(self, project_name, workflow_name):
        """
        Get a workflow by specific project name and workflow name

        :param project_name: the name of project which contains the workflow
        :param workflow_name: the workflow name
        """
        pass

    def get_workflow_by_id(self, workflow_id):
        """
        Get a workflow by specific uuid

        :param workflow_id: the uuid of workflow
        """
        pass

    def list_workflows(self, project_name):
        """
        List all workflows of the specific project

        :param project_name: the name of project which contains the workflow
        :param page_size     limitation of listed workflows.
        :param offset        offset of listed workflows.
        """
        pass

    def delete_workflow_by_name(self, project_name, workflow_name):
        """
        Delete the workflow by specific project and workflow name

        :param project_name: the name of project which contains the workflow
        :param workflow_name: the workflow name
        """
        pass

    def delete_workflow_by_id(self, workflow_id):
        """
        Delete the workflow by specific id

        :param workflow_id: the uuid of workflow
        """
        pass

    def update_workflow(self, workflow_name, project_name, properties=None):
        """
        Update the workflow

        :param workflow_name: the workflow name
        :param project_name: the name of project which contains the workflow
        :param properties: (Optional) the properties need to be updated
        """
        pass

    '''
        artifact api
    '''

    def get_artifact_by_id(self, artifact_id):
        """
        get a specific artifact in metadata store by artifact id.

        :param artifact_id: the artifact id
        :return: A single :py:class:`ai_flow.meta.artifact_meta.ArtifactMeta` object
                 if the artifact exists, Otherwise, returns None if the artifact does not exist.
        """

    def get_artifact_by_name(self, artifact_name):
        """
        get a specific artifact in metadata store by artifact name.

        :param artifact_name: the artifact name
        :return: A single :py:class:`ai_flow.meta.artifact_meta.ArtifactMeta` object
                 if the artifact exists, Otherwise, returns None if the artifact does not exist.
        """

    def register_artifact(self, name: Text, artifact_type, description,
                          uri, properties):
        """
        register an artifact in metadata store.

        :param name: the name of the artifact
        :param artifact_type: the type of the artifact
        :param description: the description of the artifact
        :param uri: the uri of the artifact
        :param properties: the properties of the artifact
        :return: A single :py:class:`ai_flow.meta.artifact_meta.py.ArtifactMeta` object.
        """

    def update_artifact(self, name: Text, artifact_type: Text, description: Text,
                        uri, properties) -> Optional[ArtifactMeta]:
        """
        Update an artifact in metadata store.

        :param name: the name of the artifact
        :param artifact_type: the type of the artifact
        :param description: the description of the artifact
        :param uri: the uri of the artifact
        :param properties: the properties of the artifact
        :return: A single :py:class:`ai_flow.meta.artifact_meta.py.ArtifactMeta` object.
        """

    def list_artifact(self, page_size, offset):
        """
        List registered artifacts in metadata store.

        :param page_size: the limitation of the listed artifacts.
        :param offset: the offset of listed artifacts.
        :return: List of :py:class:`ai_flow.meta.artifact_meta.py.ArtifactMeta` objects,
                 return None if no artifacts to be listed.
        """

    def delete_artifact_by_id(self, artifact_id):
        """
        Delete the registered artifact by artifact id .

        :param artifact_id: the artifact id
        :return: Status.OK if the artifact is successfully deleted,
                 Status.ERROR if the artifact does not exist otherwise.
        """

    def delete_artifact_by_name(self, artifact_name):
        """
        Delete the registered artifact by artifact name .

        :param artifact_name: the artifact name
        :return: Status.OK if the artifact is successfully deleted,
                 Status.ERROR if the artifact does not exist otherwise.
        """

    @abstractmethod
    def create_registered_model(self, model_name, model_desc=None):
        """
        Create a new registered model in model repository.

        :param model_name: Name of registered model. This is expected to be unique in the backend store.
        :param model_desc: (Optional) Description of registered model.

        :return: A single object of :py:class:`ai_flow.model_center.entity.RegisteredModel` created in model
        repository.
        """
        pass

    @abstractmethod
    def update_registered_model(self, registered_model, model_name=None, model_desc=None):
        """
        Update metadata for RegisteredModel entity. Either ``model_name`` or ``model_desc``
        should be non-None. Backend raises exception if a registered model with given name does not exist.

        :param registered_model: :py:class:`ai_flow.model_center.entity.RegisteredModel` object.
        :param model_name: (Optional) New proposed name for the registered model.
        :param model_desc: (Optional) Description of registered model.

        :return: A single updated :py:class:`ai_flow.model_center.entity.RegisteredModel` object.
        """
        pass

    @abstractmethod
    def delete_registered_model(self, registered_model):
        """
        Delete registered model.
        Backend raises exception if a registered model with given name does not exist.

        :param registered_model: :py:class:`ai_flow.model_center.entity.RegisteredModel` object.

        :return: None
        """
        pass

    @abstractmethod
    def list_registered_models(self):
        """
        List of all registered models in model repository.

        :return: List of :py:class:`ai_flow.model_center.entity.RegisteredModel` objects.
        """
        pass

    @abstractmethod
    def get_registered_model_detail(self, registered_model):
        """
        :param registered_model: :py:class:`ai_flow.model_center.entity.RegisteredModel` object.

        :return: A single :py:class:`ai_flow.model_center.entity.RegisteredModelDetail` object.
        """
        pass

    @abstractmethod
    def create_model_version(self, model_name, model_version, model_path,
                             model_type=None, version_desc=None):
        """
        Create a new model version from given model source and model metric.

        :param model_name: Name for containing registered model.
        :param model_version: User-defined version of registered model.
        :param model_path: Source path where the AIFlow model is stored.
        :param model_type: (Optional) Type of AIFlow model option.
        :param version_desc: (Optional) Description of registered model version.

        :return: A single object of :py:class:`ai_flow.model_center.entity.ModelVersion`
        created in model repository.
        """
        pass

    @abstractmethod
    def update_model_version(self, model_version, model_path=None, model_type=None,
                             version_desc=None, version_stage=None):
        """
        Update metadata associated with a model version in model repository.

        :param model_version: :py:class:`ai_flow.model_center.entity.ModelVersion` object.
        :param model_path: (Optional) New Source path where AIFlow model is stored.
        :param model_type: (Optional) Type of AIFlow registered model option.
        :param version_desc: (Optional) New Description of registered model version.
        :param version_stage: (Optional) New desired stage for this model version.

        :return: A single updated :py:class:`ai_flow.model_center.entity.ModelVersion` object.
        """
        pass

    @abstractmethod
    def delete_model_version(self, model_version):
        """
        Delete model version in model repository.

        :param model_version: :py:class:`ai_flow.model_center.entity.ModelVersion` object.

        :return: None
        """
        pass

    @abstractmethod
    def get_model_version_detail(self, model_version):
        """
        :param model_version: :py:class:`ai_flow.model_center.entity.ModelVersion` object.

        :return: A single :py:class:`ai_flow.model_center.entity.ModelVersionDetail` object.
        """
        pass

    @abstractmethod
    def register_metric_meta(self, metric_name, metric_type, project_name, metric_desc=None, dataset_name=None,
                             model_name=None, job_name=None, start_time=None, end_time=None, uri=None, tags=None,
                             properties=None) -> MetricMeta:
        pass

    @abstractmethod
    def update_metric_meta(self, metric_name, metric_desc=None, project_name=None, dataset_name=None,
                           model_name=None, job_name=None, start_time=None, end_time=None, uri=None, tags=None,
                           properties=None) -> MetricMeta:
        pass

    @abstractmethod
    def delete_metric_meta(self, metric_name):
        pass

    @abstractmethod
    def get_metric_meta(self, metric_name) -> Union[None, MetricMeta]:
        pass

    @abstractmethod
    def list_dataset_metric_metas(self, dataset_name, project_name=None) -> Union[None, MetricMeta, List[MetricMeta]]:
        pass

    @abstractmethod
    def list_model_metric_metas(self, model_name, project_name=None) -> Union[None, MetricMeta, List[MetricMeta]]:
        pass

    @abstractmethod
    def register_metric_summary(self, metric_name, metric_key, metric_value, metric_timestamp, model_version=None,
                                job_execution_id=None) -> MetricSummary:
        pass

    @abstractmethod
    def update_metric_summary(self, uuid, metric_name=None, metric_key=None, metric_value=None, metric_timestamp=None,
                              model_version=None, job_execution_id=None) -> MetricSummary:
        pass

    @abstractmethod
    def delete_metric_summary(self, uuid: int):
        pass

    @abstractmethod
    def get_metric_summary(self, uuid) -> Union[None, MetricSummary]:
        pass

    @abstractmethod
    def list_metric_summaries(self, metric_name=None, metric_key=None, model_version=None, start_time=None,
                              end_time=None) -> Union[None, MetricSummary, List[MetricSummary]]:
        pass

    """For high availability:"""

    @abstractmethod
    def list_living_members(self, ttl_ms) -> List['Member']:
        pass

    @abstractmethod
    def update_member(self, server_uri, server_uuid):
        pass

    @abstractmethod
    def clear_dead_members(self, ttl_ms):
        pass



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from typing import Text

from ai_flow.graph.edge import Edge


class DataEdge(Edge):
    """
    Connect two AINodes in a job.
    Data flows from the source node to the destination node.
    """
    def __init__(self,
                 source: Text,
                 destination: Text,
                 port: int = 0) -> None:
        """
        :param source: node_id of the node sending data.
        :param destination: node_id of the node receiving data.
        :param port: The serial number of the data sending port.
                     It shows the index of :class:`~ai_flow.graph.channel.Channel` in the source node.
        """
        super().__init__(destination=destination, source=source)
        self.port = port

    def __eq__(self, o: object) -> bool:
        if isinstance(o, DataEdge):
            return self.destination == o.destination \
                   and self.source == o.source \
                   and self.port == o.port
        else:
            return False

    def __ne__(self, o: object) -> bool:
        return not self.__eq__(o)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""Example DAG demonstrating the usage of the PythonOperator."""
import time
from pprint import pprint

from airflow import DAG
from airflow.operators.python import PythonOperator, PythonVirtualenvOperator
from airflow.utils.dates import days_ago

args = {
    'owner': 'airflow',
}

dag = DAG(
    dag_id='example_python_operator',
    default_args=args,
    schedule_interval=None,
    start_date=days_ago(2),
    tags=['example'],
)


# [START howto_operator_python]
def print_context(ds, **kwargs):
    """Print the Airflow context and ds variable from the context."""
    pprint(kwargs)
    print(ds)
    return 'Whatever you return gets printed in the logs'


run_this = PythonOperator(
    task_id='print_the_context',
    python_callable=print_context,
    dag=dag,
)
# [END howto_operator_python]


# [START howto_operator_python_kwargs]
def my_sleeping_function(random_base):
    """This is a function that will run within the DAG execution"""
    time.sleep(random_base)


# Generate 5 sleeping tasks, sleeping from 0.0 to 0.4 seconds respectively
for i in range(5):
    task = PythonOperator(
        task_id='sleep_for_' + str(i),
        python_callable=my_sleeping_function,
        op_kwargs={'random_base': float(i) / 10},
        dag=dag,
    )

    run_this >> task
# [END howto_operator_python_kwargs]


# [START howto_operator_python_venv]
def callable_virtualenv():
    """
    Example function that will be performed in a virtual environment.

    Importing at the module level ensures that it will not attempt to import the
    library before it is installed.
    """
    from time import sleep

    from colorama import Back, Fore, Style

    print(Fore.RED + 'some red text')
    print(Back.GREEN + 'and with a green background')
    print(Style.DIM + 'and in dim text')
    print(Style.RESET_ALL)
    for _ in range(10):
        print(Style.DIM + 'Please wait...', flush=True)
        sleep(10)
    print('Finished')


virtualenv_task = PythonVirtualenvOperator(
    task_id="virtualenv_python",
    python_callable=callable_virtualenv,
    requirements=["colorama==0.4.0"],
    system_site_packages=False,
    dag=dag,
)
# [END howto_operator_python_venv]



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from airflow.sensors.base import BaseSensorOperator
from airflow.utils import timezone
from airflow.utils.decorators import apply_defaults


class TimeDeltaSensor(BaseSensorOperator):
    """
    Waits for a timedelta after the task's execution_date + schedule_interval.
    In Airflow, the daily task stamped with ``execution_date``
    2016-01-01 can only start running on 2016-01-02. The timedelta here
    represents the time after the execution period has closed.

    :param delta: time length to wait after execution_date before succeeding
    :type delta: datetime.timedelta
    """

    @apply_defaults
    def __init__(self, *, delta, **kwargs):
        super().__init__(**kwargs)
        self.delta = delta

    def poke(self, context):
        dag = context['dag']
        target_dttm = dag.following_schedule(context['execution_date'])
        target_dttm += self.delta
        self.log.info('Checking if the time (%s) has come', target_dttm)
        return timezone.utcnow() > target_dttm



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""Base executor - this is the base class for all the implemented executors."""
import sys
import time
from collections import OrderedDict
from typing import Any, Dict, List, Optional, Set, Tuple

from airflow.events.scheduler_events import TaskStateChangedEvent
from sqlalchemy.orm import Session, selectinload

from airflow.configuration import conf
from airflow.executors.scheduling_action import SchedulingAction
from airflow.models.taskinstance import TaskInstance, TaskInstanceKey
from airflow.stats import Stats
from airflow.utils.log.logging_mixin import LoggingMixin
from airflow.utils.session import provide_session, create_session
from airflow.utils.state import State
from airflow.settings import CHECK_TASK_STOPPED_INTERVAL, CHECK_TASK_STOPPED_NUM

PARALLELISM: int = conf.getint('core', 'PARALLELISM')

NOT_STARTED_MESSAGE = "The executor should be started first!"

# Command to execute - list of strings
# the first element is always "airflow".
# It should be result of TaskInstance.generate_command method.q
CommandType = List[str]


# Task that is queued. It contains all the information that is
# needed to run the task.
#
# Tuple of: command, priority, queue name, TaskInstance
QueuedTaskInstanceType = Tuple[CommandType, int, Optional[str], TaskInstance]

# Event_buffer dict value type
# Tuple of: state, info
EventBufferValueType = Tuple[Optional[str], Any]


class BaseExecutor(LoggingMixin):
    """
    Class to derive in order to interface with executor-type systems
    like Celery, Kubernetes, Local, Sequential and the likes.

    :param parallelism: how many jobs should run at one time. Set to
        ``0`` for infinity
    """

    job_id: Optional[str] = None

    def __init__(self, parallelism: int = PARALLELISM):
        super().__init__()
        self.parallelism: int = parallelism
        self.queued_tasks: OrderedDict[TaskInstanceKey, QueuedTaskInstanceType] = OrderedDict()
        self.running: Set[TaskInstanceKey] = set()
        self.event_buffer: Dict[TaskInstanceKey, EventBufferValueType] = {}
        self._mailbox = None
        self._server_uri = None

    def set_mailbox(self, mailbox):
        self._mailbox = mailbox

    def set_server_uri(self, server_uri):
        self._server_uri = server_uri

    def start(self):  # pragma: no cover
        """Executors may need to get things started."""

    def queue_command(
        self,
        task_instance: TaskInstance,
        command: CommandType,
        priority: int = 1,
        queue: Optional[str] = None,
    ):
        """Queues command to task"""
        if task_instance.key not in self.queued_tasks and task_instance.key not in self.running:
            self.log.info("Adding to queue: %s", command)
            self.queued_tasks[task_instance.key] = (command, priority, queue, task_instance)
        else:
            self.log.error("could not queue task %s", task_instance.key)

    def queue_task_instance(
        self,
        task_instance: TaskInstance,
        mark_success: bool = False,
        pickle_id: Optional[str] = None,
        ignore_all_deps: bool = False,
        ignore_depends_on_past: bool = False,
        ignore_task_deps: bool = False,
        ignore_ti_state: bool = False,
        pool: Optional[str] = None,
        cfg_path: Optional[str] = None,
    ) -> None:
        """Queues task instance."""
        pool = pool or task_instance.pool

        # TODO (edgarRd): AIRFLOW-1985:
        # cfg_path is needed to propagate the config values if using impersonation
        # (run_as_user), given that there are different code paths running tasks.
        # For a long term solution we need to address AIRFLOW-1986
        command_list_to_run = task_instance.command_as_list(
            local=True,
            mark_success=mark_success,
            ignore_all_deps=ignore_all_deps,
            ignore_depends_on_past=ignore_depends_on_past,
            ignore_task_deps=ignore_task_deps,
            ignore_ti_state=ignore_ti_state,
            pool=pool,
            pickle_id=pickle_id,
            cfg_path=cfg_path,
        )
        self.log.debug("created command %s", command_list_to_run)
        self.queue_command(
            task_instance,
            command_list_to_run,
            priority=task_instance.task.priority_weight_total,
            queue=task_instance.task.queue,
        )

    def schedule_task(self, key: TaskInstanceKey, action: SchedulingAction):
        """
        Schedule a task

        :param key: task instance key
        :param action: task scheduling action in [START, STOP, RESTART]
        """
        with create_session() as session:
            ti = session.query(TaskInstance).filter(
                TaskInstance.dag_id == key.dag_id,
                TaskInstance.task_id == key.task_id,
                TaskInstance.execution_date == key.execution_date
            ).first()
            if SchedulingAction.START == action:
                if ti.state not in State.running:
                    self._start_task_instance(key)
                    self._send_message(ti)
            elif SchedulingAction.STOP == action:
                if ti.state in State.unfinished:
                    if self._stop_task_instance(key):
                        self._send_message(ti)
            elif SchedulingAction.RESTART == action:
                if ti.state in State.running:
                    self._restart_task_instance(key)
                else:
                    self._start_task_instance(key)
                self._send_message(ti)
            else:
                raise ValueError('The task scheduling action must in ["START", "STOP", "RESTART"].')

    def _send_message(self, ti):
        self.send_message(TaskInstanceKey(ti.dag_id, ti.task_id, ti.execution_date, ti.try_number))

    def _start_task_instance(self, key: TaskInstanceKey):
        """
        Ignore all dependencies, force start a task instance
        """
        ti = self.get_task_instance(key)
        if ti is None:
            self.log.error("TaskInstance not found in DB, %s.", str(key))
            return
        command = TaskInstance.generate_command(
            ti.dag_id,
            ti.task_id,
            ti.execution_date,
            local=True,
            mark_success=False,
            ignore_all_deps=True,
            ignore_depends_on_past=True,
            ignore_task_deps=True,
            ignore_ti_state=True,
            pool=ti.pool,
            file_path=ti.dag_model.fileloc,
            pickle_id=ti.dag_model.pickle_id,
            server_uri=self._server_uri,
        )
        ti.set_state(State.QUEUED)
        self.execute_async(
            key=key,
            command=command,
            queue=ti.queue,
            executor_config=ti.executor_config
        )

    def _stop_related_process(self, ti: TaskInstance) -> bool:
        raise NotImplementedError()

    def _stop_task_instance(self, key: TaskInstanceKey) -> bool:
        """
        Force stopping a task instance.
        return: True if successfully kill the task instance
        """
        ti = self.get_task_instance(key)
        if ti is None:
            self.log.error("Task not found in DB, %s.", str(key))
            return False
        elif ti.state in State.finished:
            self.log.info("Task has been finished with state %s, %s", ti.state, str(key))
            return True
        elif ti.state != State.RUNNING:
            ti.set_state(State.KILLED)
            return True
        try:
            self.running.remove(ti.key)
        except KeyError:
            self.log.debug('Could not find key: %s', str(ti.key))
        ti.set_state(State.KILLING)
        self._stop_related_process(ti)
        stopped = self._wait_for_stopping_task(
            key=key,
            check_interval=CHECK_TASK_STOPPED_INTERVAL,
            max_check_num=CHECK_TASK_STOPPED_NUM)
        if not stopped:
            self.log.error("Failed to stop task instance: %s, please try again.", str(ti))
        return stopped

    def _restart_task_instance(self, key: TaskInstanceKey):
        """Force restarting a task instance"""
        if self._stop_task_instance(key):
            self._start_task_instance(key)

    def has_task(self, task_instance: TaskInstance) -> bool:
        """
        Checks if a task is either queued or running in this executor.

        :param task_instance: TaskInstance
        :return: True if the task is known to this executor
        """
        return task_instance.key in self.queued_tasks or task_instance.key in self.running

    def sync(self) -> None:
        """
        Sync will get called periodically by the heartbeat method.
        Executors should override this to perform gather statuses.
        """

    def heartbeat(self) -> None:
        """Heartbeat sent to trigger new jobs."""
        if not self.parallelism:
            open_slots = len(self.queued_tasks)
        else:
            open_slots = self.parallelism - len(self.running)

        num_running_tasks = len(self.running)
        num_queued_tasks = len(self.queued_tasks)

        self.log.debug("%s running task instances", num_running_tasks)
        self.log.debug("%s in queue", num_queued_tasks)
        self.log.debug("%s open slots", open_slots)

        Stats.gauge('executor.open_slots', open_slots)
        Stats.gauge('executor.queued_tasks', num_queued_tasks)
        Stats.gauge('executor.running_tasks', num_running_tasks)

        self.trigger_tasks(open_slots)

        # Calling child class sync method
        self.log.debug("Calling the %s sync method", self.__class__)
        self.sync()

    def order_queued_tasks_by_priority(self) -> List[Tuple[TaskInstanceKey, QueuedTaskInstanceType]]:
        """
        Orders the queued tasks by priority.

        :return: List of tuples from the queued_tasks according to the priority.
        """
        return sorted(
            [(k, v) for k, v in self.queued_tasks.items()],  # pylint: disable=unnecessary-comprehension
            key=lambda x: x[1][1],
            reverse=True,
        )

    def trigger_tasks(self, open_slots: int) -> None:
        """
        Triggers tasks

        :param open_slots: Number of open slots
        """
        sorted_queue = self.order_queued_tasks_by_priority()

        for _ in range(min((open_slots, len(self.queued_tasks)))):
            key, (command, _, _, ti) = sorted_queue.pop(0)
            self.queued_tasks.pop(key)
            self.running.add(key)
            self.execute_async(key=key, command=command, queue=None, executor_config=ti.executor_config)

    def change_state(self, key: TaskInstanceKey, state: str, info=None) -> None:
        """
        Changes state of the task.

        :param info: Executor information for the task instance
        :param key: Unique key for the task instance
        :param state: State to set for the task.
        """
        self.log.debug("Changing state: %s", key)
        try:
            self.running.remove(key)
        except KeyError:
            self.log.debug('Could not find key: %s', str(key))
        self.event_buffer[key] = state, info

    def fail(self, key: TaskInstanceKey, info=None) -> None:
        """
        Set fail state for the event.

        :param info: Executor information for the task instance
        :param key: Unique key for the task instance
        """
        self.change_state(key, State.FAILED, info)

    def success(self, key: TaskInstanceKey, info=None) -> None:
        """
        Set success state for the event.

        :param info: Executor information for the task instance
        :param key: Unique key for the task instance
        """
        self.change_state(key, State.SUCCESS, info)

    def get_event_buffer(self, dag_ids=None) -> Dict[TaskInstanceKey, EventBufferValueType]:
        """
        Returns and flush the event buffer. In case dag_ids is specified
        it will only return and flush events for the given dag_ids. Otherwise
        it returns and flushes all events.

        :param dag_ids: to dag_ids to return events for, if None returns all
        :return: a dict of events
        """
        cleared_events: Dict[TaskInstanceKey, EventBufferValueType] = {}
        if dag_ids is None:
            cleared_events = self.event_buffer
            self.event_buffer = {}
        else:
            for ti_key in list(self.event_buffer.keys()):
                if ti_key.dag_id in dag_ids:
                    cleared_events[ti_key] = self.event_buffer.pop(ti_key)

        return cleared_events

    def execute_async(
        self,
        key: TaskInstanceKey,
        command: CommandType,
        queue: Optional[str] = None,
        executor_config: Optional[Any] = None,
    ) -> None:  # pragma: no cover
        """
        This method will execute the command asynchronously.

        :param key: Unique key for the task instance
        :param command: Command to run
        :param queue: name of the queue
        :param executor_config: Configuration passed to the executor.
        """
        raise NotImplementedError()

    def end(self) -> None:  # pragma: no cover
        """
        This method is called when the caller is done submitting job and
        wants to wait synchronously for the job submitted previously to be
        all done.
        """
        raise NotImplementedError()

    def terminate(self):
        """This method is called when the daemon receives a SIGTERM"""
        raise NotImplementedError()

    def try_adopt_task_instances(self, tis: List[TaskInstance]) -> List[TaskInstance]:
        """
        Try to adopt running task instances that have been abandoned by a SchedulerJob dying.

        Anything that is not adopted will be cleared by the scheduler (and then become eligible for
        re-scheduling)

        :return: any TaskInstances that were unable to be adopted
        :rtype: list[airflow.models.TaskInstance]
        """
        # By default, assume Executors cannot adopt tasks, so just say we failed to adopt anything.
        # Subclasses can do better!
        return tis

    @property
    def slots_available(self):
        """Number of new tasks this executor instance can accept"""
        if self.parallelism:
            return self.parallelism - len(self.running) - len(self.queued_tasks)
        else:
            return sys.maxsize

    @staticmethod
    def validate_command(command: List[str]) -> None:
        """Check if the command to execute is airflow command"""
        if command[0:3] != ["airflow", "tasks", "run"]:
            raise ValueError('The command must start with ["airflow", "tasks", "run"].')

    def debug_dump(self):
        """Called in response to SIGUSR2 by the scheduler"""
        self.log.info(
            "executor.queued (%d)\n\t%s",
            len(self.queued_tasks),
            "\n\t".join(map(repr, self.queued_tasks.items())),
        )
        self.log.info("executor.running (%d)\n\t%s", len(self.running), "\n\t".join(map(repr, self.running)))
        self.log.info(
            "executor.event_buffer (%d)\n\t%s",
            len(self.event_buffer),
            "\n\t".join(map(repr, self.event_buffer.items())),
        )

    @provide_session
    def get_task_instance(self, key: TaskInstanceKey, session: Session = None) -> Optional[TaskInstance]:
        """
        Returns the task instance specified by TaskInstanceKey

        :param key: the task instance key
        :type key: TaskInstanceKey
        :param session: Sqlalchemy ORM Session
        :type session: Session
        """
        return (
            session.query(TaskInstance).filter(
                TaskInstance.dag_id == key.dag_id,
                TaskInstance.execution_date == key.execution_date,
                TaskInstance.task_id == key.task_id
            ).options(selectinload('dag_model')).first()
        )

    def send_message(self, key: TaskInstanceKey):
        if self._mailbox is not None:
            ti = self.get_task_instance(key)
            task_status_changed_event = TaskStateChangedEvent(
                ti.task_id,
                ti.dag_id,
                ti.execution_date,
                ti.state
            )
            self._mailbox.send_message(task_status_changed_event.to_event())

    def _wait_for_stopping_task(
        self,
        key: TaskInstanceKey,
        check_interval: int,
        max_check_num: int,
    ) -> bool:
        """
        Wait check_interval * max_check_num seconds for task to be killed successfully
        :param key: the task instance key
        :type key: TaskInstanceKey
        :param check_interval: interval between checking state
        :type check_interval: int
        :param max_check_num: max times of checking state
        :type max_check_num: int
        """
        check_num = 0
        while check_num < max_check_num and check_interval > 0:
            check_num = check_num + 1
            ti = self.get_task_instance(key)
            if ti and ti.state == State.KILLED:
                return True
            else:
                time.sleep(check_interval)
        return False

    def recover_state(self):
        """
        Recover the state of dags after restarting scheduler.
        """
        pass



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
import logging
import os
import re
import urllib.parse
from contextlib import contextmanager

import sqlalchemy

from ai_flow.protobuf.message_pb2 import INTERNAL_ERROR
from ai_flow.endpoint.server.exception import AIFlowException
from ai_flow.store import AIFLOW_SQLALCHEMYSTORE_MAX_OVERFLOW, AIFLOW_SQLALCHEMYSTORE_POOL_SIZE
from ai_flow.store.db.db_engine import DATABASE_ENGINES

_logger = logging.getLogger(__name__)


def extract_db_engine_from_uri(db_uri):
    """
    Parse specified database URI to extract database type. Confirm extracted database engine is
    supported. If database driver is specified, confirm driver passes a plausible regex.
    """
    scheme = urllib.parse.urlparse(db_uri).scheme
    scheme_plus_count = scheme.count('+')

    """validates scheme parsed from DB URI is supported"""
    if scheme_plus_count == 0:
        db_engine = scheme
    elif scheme_plus_count == 1:
        db_engine, _ = scheme.split('+')
    else:
        error_msg = "Invalid database URI: '%s'." % db_uri
        raise AIFlowException(error_msg)

    """validates db_engine parsed from DB URI is supported"""
    if db_engine not in DATABASE_ENGINES:
        error_msg = "Invalid database engine: '%s'." % db_engine
        raise AIFlowException(error_msg)

    return db_engine


def _get_managed_session_maker(SessionMaker):
    """
    Creates session factory for generating exception-safe SQLAlchemy sessions that are available for
    using session context manager. Session generated by session factory is automatically committed
    if no exceptions are encountered within its associated context. If an exception is
    encountered, this session could be rolled back. Session generated by session factory is
    automatically closed when the session's associated context is exited.
    """

    @contextmanager
    def make_managed_session():
        """Provide transactional scope around series of session operations."""
        session = SessionMaker()
        try:
            yield session
            session.commit()
        except AIFlowException:
            session.rollback()
            raise
        except Exception as e:
            session.rollback()
            raise AIFlowException(error_msg=e, error_code=INTERNAL_ERROR)
        finally:
            session.close()

    return make_managed_session


def create_sqlalchemy_engine(db_uri):
    """
    Create SQLAlchemy engine with specified database URI to support AIFlow entities backend storage.
    """
    pool_size = os.environ.get(AIFLOW_SQLALCHEMYSTORE_POOL_SIZE)
    pool_max_overflow = os.environ.get(AIFLOW_SQLALCHEMYSTORE_MAX_OVERFLOW)
    pool_kwargs = {}
    if pool_size:
        pool_kwargs['pool_size'] = int(pool_size)
    if pool_max_overflow:
        pool_kwargs['max_overflow'] = int(pool_max_overflow)
    if pool_kwargs:
        _logger.info("Create SQLAlchemy engine with pool options %s", pool_kwargs)
    return sqlalchemy.create_engine(db_uri, pool_pre_ping=True, **pool_kwargs)

def parse_mongo_uri(db_uri):
    """
    Parse MongoDB URI-style string to split up and return credentials

    Args:
        db_uri (string): MongoDB URI-style string
    Return:

    """
    regex_str = r'^(?P<schema>(mongodb:(?:\/{2})?))((?P<user>\w+?):(?P<pwd>(\w+?))@|:@?)(?P<host>(\S+?)):(?P<port>(\d+))(\/(?P<db>(\S+?)))$'
    pattern = re.compile(regex_str)
    m = pattern.match(db_uri)
    if m is None:
        raise Exception('The URI of MongoDB is invalid')
    return m.group('user'), m.group('pwd'), m.group('host'), m.group('port'), m.group('db')


# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""Webserver command"""
import hashlib
import logging
import os
import signal
import subprocess
import sys
import textwrap
import time
from contextlib import suppress
from time import sleep
from typing import Dict, List, NoReturn

import daemon
import psutil
from daemon.pidfile import TimeoutPIDLockFile
from lockfile.pidlockfile import read_pid_from_pidfile

from airflow import settings
from airflow.configuration import conf
from airflow.exceptions import AirflowException, AirflowWebServerTimeout
from airflow.utils import cli as cli_utils
from airflow.utils.cli import setup_locations, setup_logging
from airflow.utils.log.logging_mixin import LoggingMixin
from airflow.utils.process_utils import check_if_pidfile_process_is_running
from airflow.www.app import cached_app, create_app

log = logging.getLogger(__name__)


class GunicornMonitor(LoggingMixin):
    """
    Runs forever, monitoring the child processes of @gunicorn_master_proc and
    restarting workers occasionally or when files in the plug-in directory
    has been modified.

    Each iteration of the loop traverses one edge of this state transition
    diagram, where each state (node) represents
    [ num_ready_workers_running / num_workers_running ]. We expect most time to
    be spent in [n / n]. `bs` is the setting webserver.worker_refresh_batch_size.
    The horizontal transition at ? happens after the new worker parses all the
    dags (so it could take a while!)
       V ────────────────────────────────────────────────────────────────────────┐
    [n / n] ──TTIN──> [ [n, n+bs) / n + bs ]  ────?───> [n + bs / n + bs] ──TTOU─┘
       ^                          ^───────────────┘
       │
       │      ┌────────────────v
       └──────┴────── [ [0, n) / n ] <─── start
    We change the number of workers by sending TTIN and TTOU to the gunicorn
    master process, which increases and decreases the number of child workers
    respectively. Gunicorn guarantees that on TTOU workers are terminated
    gracefully and that the oldest worker is terminated.

    :param gunicorn_master_pid: PID for the main Gunicorn process
    :param num_workers_expected: Number of workers to run the Gunicorn web server
    :param master_timeout: Number of seconds the webserver waits before killing gunicorn master that
        doesn't respond
    :param worker_refresh_interval: Number of seconds to wait before refreshing a batch of workers.
    :param worker_refresh_batch_size: Number of workers to refresh at a time. When set to 0, worker
        refresh is disabled. When nonzero, airflow periodically refreshes webserver workers by
        bringing up new ones and killing old ones.
    :param reload_on_plugin_change: If set to True, Airflow will track files in plugins_folder directory.
        When it detects changes, then reload the gunicorn.
    """

    def __init__(
        self,
        gunicorn_master_pid: int,
        num_workers_expected: int,
        master_timeout: int,
        worker_refresh_interval: int,
        worker_refresh_batch_size: int,
        reload_on_plugin_change: bool,
    ):
        super().__init__()
        self.gunicorn_master_proc = psutil.Process(gunicorn_master_pid)
        self.num_workers_expected = num_workers_expected
        self.master_timeout = master_timeout
        self.worker_refresh_interval = worker_refresh_interval
        self.worker_refresh_batch_size = worker_refresh_batch_size
        self.reload_on_plugin_change = reload_on_plugin_change

        self._num_workers_running = 0
        self._num_ready_workers_running = 0
        self._last_refresh_time = time.monotonic() if worker_refresh_interval > 0 else None
        self._last_plugin_state = self._generate_plugin_state() if reload_on_plugin_change else None
        self._restart_on_next_plugin_check = False

    def _generate_plugin_state(self) -> Dict[str, float]:
        """
        Generate dict of filenames and last modification time of all files in settings.PLUGINS_FOLDER
        directory.
        """
        if not settings.PLUGINS_FOLDER:
            return {}

        all_filenames: List[str] = []
        for (root, _, filenames) in os.walk(settings.PLUGINS_FOLDER):
            all_filenames.extend(os.path.join(root, f) for f in filenames)
        plugin_state = {f: self._get_file_hash(f) for f in sorted(all_filenames)}
        return plugin_state

    @staticmethod
    def _get_file_hash(fname: str):
        """Calculate MD5 hash for file"""
        hash_md5 = hashlib.md5()
        with open(fname, "rb") as f:
            for chunk in iter(lambda: f.read(4096), b""):
                hash_md5.update(chunk)
        return hash_md5.hexdigest()

    def _get_num_ready_workers_running(self) -> int:
        """Returns number of ready Gunicorn workers by looking for READY_PREFIX in process name"""
        workers = psutil.Process(self.gunicorn_master_proc.pid).children()

        def ready_prefix_on_cmdline(proc):
            try:
                cmdline = proc.cmdline()
                if len(cmdline) > 0:  # pylint: disable=len-as-condition
                    return settings.GUNICORN_WORKER_READY_PREFIX in cmdline[0]
            except psutil.NoSuchProcess:
                pass
            return False

        ready_workers = [proc for proc in workers if ready_prefix_on_cmdline(proc)]
        return len(ready_workers)

    def _get_num_workers_running(self) -> int:
        """Returns number of running Gunicorn workers processes"""
        workers = psutil.Process(self.gunicorn_master_proc.pid).children()
        return len(workers)

    def _wait_until_true(self, fn, timeout: int = 0) -> None:
        """Sleeps until fn is true"""
        start_time = time.monotonic()
        while not fn():
            if 0 < timeout <= time.monotonic() - start_time:
                raise AirflowWebServerTimeout(f"No response from gunicorn master within {timeout} seconds")
            sleep(0.1)

    def _spawn_new_workers(self, count: int) -> None:
        """
        Send signal to kill the worker.

        :param count: The number of workers to spawn
        """
        excess = 0
        for _ in range(count):
            # TTIN: Increment the number of processes by one
            self.gunicorn_master_proc.send_signal(signal.SIGTTIN)
            excess += 1
            self._wait_until_true(
                lambda: self.num_workers_expected + excess == self._get_num_workers_running(),
                timeout=self.master_timeout,
            )

    def _kill_old_workers(self, count: int) -> None:
        """
        Send signal to kill the worker.

        :param count: The number of workers to kill
        """
        for _ in range(count):
            count -= 1
            # TTOU: Decrement the number of processes by one
            self.gunicorn_master_proc.send_signal(signal.SIGTTOU)
            self._wait_until_true(
                lambda: self.num_workers_expected + count == self._get_num_workers_running(),
                timeout=self.master_timeout,
            )

    def _reload_gunicorn(self) -> None:
        """
        Send signal to reload the gunciron configuration. When gunciorn receive signals, it reload the
        configuration, start the new worker processes with a new configuration and gracefully
        shutdown older workers.
        """
        # HUP: Reload the configuration.
        self.gunicorn_master_proc.send_signal(signal.SIGHUP)
        sleep(1)
        self._wait_until_true(
            lambda: self.num_workers_expected == self._get_num_workers_running(), timeout=self.master_timeout
        )

    def start(self) -> NoReturn:
        """Starts monitoring the webserver."""
        try:  # pylint: disable=too-many-nested-blocks
            self._wait_until_true(
                lambda: self.num_workers_expected == self._get_num_workers_running(),
                timeout=self.master_timeout,
            )
            while True:
                if not self.gunicorn_master_proc.is_running():
                    sys.exit(1)
                self._check_workers()
                # Throttle loop
                sleep(1)

        except (AirflowWebServerTimeout, OSError) as err:
            self.log.error(err)
            self.log.error("Shutting down webserver")
            try:
                self.gunicorn_master_proc.terminate()
                self.gunicorn_master_proc.wait()
            finally:
                sys.exit(1)

    def _check_workers(self) -> None:
        num_workers_running = self._get_num_workers_running()
        num_ready_workers_running = self._get_num_ready_workers_running()

        # Whenever some workers are not ready, wait until all workers are ready
        if num_ready_workers_running < num_workers_running:
            self.log.debug(
                '[%d / %d] Some workers are starting up, waiting...',
                num_ready_workers_running,
                num_workers_running,
            )
            sleep(1)
            return

        # If there are too many workers, then kill a worker gracefully by asking gunicorn to reduce
        # number of workers
        if num_workers_running > self.num_workers_expected:
            excess = min(num_workers_running - self.num_workers_expected, self.worker_refresh_batch_size)
            self.log.debug(
                '[%d / %d] Killing %s workers', num_ready_workers_running, num_workers_running, excess
            )
            self._kill_old_workers(excess)
            return

        # If there are too few workers, start a new worker by asking gunicorn
        # to increase number of workers
        if num_workers_running < self.num_workers_expected:
            self.log.error(
                "[%d / %d] Some workers seem to have died and gunicorn did not restart them as expected",
                num_ready_workers_running,
                num_workers_running,
            )
            sleep(10)
            num_workers_running = self._get_num_workers_running()
            if num_workers_running < self.num_workers_expected:
                new_worker_count = min(
                    num_workers_running - self.worker_refresh_batch_size, self.worker_refresh_batch_size
                )
                self.log.debug(
                    '[%d / %d] Spawning %d workers',
                    num_ready_workers_running,
                    num_workers_running,
                    new_worker_count,
                )
                self._spawn_new_workers(num_workers_running)
            return

        # Now the number of running and expected worker should be equal

        # If workers should be restarted periodically.
        if self.worker_refresh_interval > 0 and self._last_refresh_time:
            # and we refreshed the workers a long time ago, refresh the workers
            last_refresh_diff = time.monotonic() - self._last_refresh_time
            if self.worker_refresh_interval < last_refresh_diff:
                num_new_workers = self.worker_refresh_batch_size
                self.log.debug(
                    '[%d / %d] Starting doing a refresh. Starting %d workers.',
                    num_ready_workers_running,
                    num_workers_running,
                    num_new_workers,
                )
                self._spawn_new_workers(num_new_workers)
                self._last_refresh_time = time.monotonic()
                return

        # if we should check the directory with the plugin,
        if self.reload_on_plugin_change:
            # compare the previous and current contents of the directory
            new_state = self._generate_plugin_state()
            # If changed, wait until its content is fully saved.
            if new_state != self._last_plugin_state:
                self.log.debug(
                    '[%d / %d] Plugins folder changed. The gunicorn will be restarted the next time the '
                    'plugin directory is checked, if there is no change in it.',
                    num_ready_workers_running,
                    num_workers_running,
                )
                self._restart_on_next_plugin_check = True
                self._last_plugin_state = new_state
            elif self._restart_on_next_plugin_check:
                self.log.debug(
                    '[%d / %d] Starts reloading the gunicorn configuration.',
                    num_ready_workers_running,
                    num_workers_running,
                )
                self._restart_on_next_plugin_check = False
                self._last_refresh_time = time.monotonic()
                self._reload_gunicorn()


@cli_utils.action_logging
def webserver(args):
    """Starts Airflow Webserver"""
    print(settings.HEADER)

    # Check for old/insecure config, and fail safe (i.e. don't launch) if the config is wildly insecure.
    if conf.get('webserver', 'secret_key') == 'temporary_key':
        from rich import print as rich_print

        rich_print(
            "[red][bold]ERROR:[/bold] The `secret_key` setting under the webserver config has an insecure "
            "value - Airflow has failed safe and refuses to start. Please change this value to a new, "
            "per-environment, randomly generated string, for example using this command `[cyan]openssl rand "
            "-hex 30[/cyan]`",
            file=sys.stderr,
        )
        sys.exit(1)

    access_logfile = args.access_logfile or conf.get('webserver', 'access_logfile')
    error_logfile = args.error_logfile or conf.get('webserver', 'error_logfile')
    access_logformat = args.access_logformat or conf.get('webserver', 'access_logformat')
    num_workers = args.workers or conf.get('webserver', 'workers')
    worker_timeout = args.worker_timeout or conf.get('webserver', 'web_server_worker_timeout')
    ssl_cert = args.ssl_cert or conf.get('webserver', 'web_server_ssl_cert')
    ssl_key = args.ssl_key or conf.get('webserver', 'web_server_ssl_key')
    if not ssl_cert and ssl_key:
        raise AirflowException('An SSL certificate must also be provided for use with ' + ssl_key)
    if ssl_cert and not ssl_key:
        raise AirflowException('An SSL key must also be provided for use with ' + ssl_cert)

    if args.debug:
        print(f"Starting the web server on port {args.port} and host {args.hostname}.")
        app = create_app(testing=conf.getboolean('core', 'unit_test_mode'), server_uri=args.server_uri)
        app.run(
            debug=True,
            use_reloader=not app.config['TESTING'],
            port=args.port,
            host=args.hostname,
            ssl_context=(ssl_cert, ssl_key) if ssl_cert and ssl_key else None,
        )
    else:
        # This pre-warms the cache, and makes possible errors
        # get reported earlier (i.e. before demonization)
        os.environ['SKIP_DAGS_PARSING'] = 'True'
        app = cached_app(None)
        os.environ.pop('SKIP_DAGS_PARSING')

        pid_file, stdout, stderr, log_file = setup_locations(
            "webserver", args.pid, args.stdout, args.stderr, args.log_file
        )

        # Check if webserver is already running if not, remove old pidfile
        check_if_pidfile_process_is_running(pid_file=pid_file, process_name="webserver")

        print(
            textwrap.dedent(
                '''\
                Running the Gunicorn Server with:
                Workers: {num_workers} {workerclass}
                Host: {hostname}:{port}
                Timeout: {worker_timeout}
                Logfiles: {access_logfile} {error_logfile}
                Access Logformat: {access_logformat}
                =================================================================\
            '''.format(
                    num_workers=num_workers,
                    workerclass=args.workerclass,
                    hostname=args.hostname,
                    port=args.port,
                    worker_timeout=worker_timeout,
                    access_logfile=access_logfile,
                    error_logfile=error_logfile,
                    access_logformat=access_logformat,
                )
            )
        )

        run_args = [
            'gunicorn',
            '--workers',
            str(num_workers),
            '--worker-class',
            str(args.workerclass),
            '--timeout',
            str(worker_timeout),
            '--bind',
            args.hostname + ':' + str(args.port),
            '--name',
            'airflow-webserver',
            '--pid',
            pid_file,
            '--config',
            'python:airflow.www.gunicorn_config',
        ]

        if args.access_logfile:
            run_args += ['--access-logfile', str(args.access_logfile)]

        if args.error_logfile:
            run_args += ['--error-logfile', str(args.error_logfile)]

        if args.access_logformat and args.access_logformat.strip():
            run_args += ['--access-logformat', str(args.access_logformat)]

        if args.daemon:
            run_args += ['--daemon']

        if ssl_cert:
            run_args += ['--certfile', ssl_cert, '--keyfile', ssl_key]

        run_args += ["airflow.www.app:cached_app()"]

        gunicorn_master_proc = None

        def kill_proc(signum, _):  # pylint: disable=unused-argument
            log.info("Received signal: %s. Closing gunicorn.", signum)
            gunicorn_master_proc.terminate()
            with suppress(TimeoutError):
                gunicorn_master_proc.wait(timeout=30)
            if gunicorn_master_proc.poll() is not None:
                gunicorn_master_proc.kill()
            sys.exit(0)

        def monitor_gunicorn(gunicorn_master_pid: int):
            # Register signal handlers
            signal.signal(signal.SIGINT, kill_proc)
            signal.signal(signal.SIGTERM, kill_proc)

            # These run forever until SIG{INT, TERM, KILL, ...} signal is sent
            GunicornMonitor(
                gunicorn_master_pid=gunicorn_master_pid,
                num_workers_expected=num_workers,
                master_timeout=conf.getint('webserver', 'web_server_master_timeout'),
                worker_refresh_interval=conf.getint('webserver', 'worker_refresh_interval', fallback=30),
                worker_refresh_batch_size=conf.getint('webserver', 'worker_refresh_batch_size', fallback=1),
                reload_on_plugin_change=conf.getboolean(
                    'webserver', 'reload_on_plugin_change', fallback=False
                ),
            ).start()

        os.environ['SERVER_URI'] = args.server_uri

        if args.daemon:
            handle = setup_logging(log_file)

            base, ext = os.path.splitext(pid_file)
            with open(stdout, 'w+') as stdout, open(stderr, 'w+') as stderr:
                ctx = daemon.DaemonContext(
                    pidfile=TimeoutPIDLockFile(f"{base}-monitor{ext}", -1),
                    files_preserve=[handle],
                    stdout=stdout,
                    stderr=stderr,
                )
                with ctx:
                    subprocess.Popen(run_args, close_fds=True, env=os.environ)

                    # Reading pid of gunicorn master as it will be different that
                    # the one of process spawned above.
                    while True:
                        sleep(0.1)
                        gunicorn_master_proc_pid = read_pid_from_pidfile(pid_file)
                        if gunicorn_master_proc_pid:
                            break

                    # Run Gunicorn monitor
                    gunicorn_master_proc = psutil.Process(gunicorn_master_proc_pid)
                    monitor_gunicorn(gunicorn_master_proc.pid)

        else:
            gunicorn_master_proc = subprocess.Popen(run_args, close_fds=True, env=os.environ)
            monitor_gunicorn(gunicorn_master_proc.pid)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
import unittest

from ai_flow.util import json_utils

from ai_flow.workflow.control_edge import EventMeetConfig, EventLife, ValueCondition, EventCondition, ConditionType, \
    MeetAnyEventCondition, MeetAllEventCondition, SchedulingRule, JobAction, ControlEdge
from notification_service.base_notification import UNDEFINED_EVENT_TYPE, DEFAULT_NAMESPACE


class TestEventMeetConfig(unittest.TestCase):

    def test_event_meet_config(self):
        event_config = EventMeetConfig(event_key='key', event_value='value')
        self.assertEqual(event_config.event_key, 'key')
        self.assertEqual(event_config.event_value, 'value')
        self.assertEqual(event_config.event_type, UNDEFINED_EVENT_TYPE)
        self.assertEqual(event_config.namespace, DEFAULT_NAMESPACE)
        self.assertIsNone(event_config.sender)
        self.assertEqual(event_config.life, EventLife.ONCE)
        self.assertEqual(event_config.value_condition, ValueCondition.EQUALS)


class TestEventCondition(unittest.TestCase):

    def test_event_condition_add_event(self):
        condition = EventCondition([], ConditionType.MEET_ANY)
        condition.add_event(event_key='k1', event_value='v1')
        condition.add_event(event_key='k2', event_value='v2')
        expected_list = [EventMeetConfig('k1', 'v1'), EventMeetConfig('k2', 'v2')]
        self.assertListEqual(expected_list, condition.events)

        # duplicate event will not added twice
        condition.add_event(event_key='k1', event_value='v1')
        condition.add_event(event_key='k2', event_value='v2')
        self.assertListEqual(expected_list, condition.events)

    def test_event_condition_serde(self):
        condition = EventCondition([], ConditionType.MEET_ANY)
        condition.add_event(event_key='k1', event_value='v1')
        condition.add_event(event_key='k2', event_value='v2')
        loaded_condition = json_utils.loads(json_utils.dumps(condition))
        self.assertEqual(condition, loaded_condition)
        loaded_condition.add_event(event_key='k1', event_value='v1')
        loaded_condition.add_event(event_key='k2', event_value='v2')
        self.assertEqual(condition, loaded_condition)


class MeetAnyEventConditionTest(unittest.TestCase):

    def test_meet_any_event_condition(self):
        condition = MeetAnyEventCondition()
        self.assertEqual(ConditionType.MEET_ANY, condition.condition_type)


class MeetAllEventConditionTest(unittest.TestCase):

    def test_meet_all_event_condition(self):
        condition = MeetAllEventCondition()
        self.assertEqual(ConditionType.MEET_ALL, condition.condition_type)


class SchedulingRuleTest(unittest.TestCase):

    def test_scheduling_rule(self):
        condition = MeetAllEventCondition()
        action = JobAction.START
        rule = SchedulingRule(condition, action)
        self.assertEqual(condition, rule.event_condition)
        self.assertEqual(action, rule.action)

    def test_scheduling_rule_serde(self):
        condition = MeetAllEventCondition()
        action = JobAction.START
        rule = SchedulingRule(condition, action)
        loaded_rule = json_utils.loads(json_utils.dumps(rule))
        self.assertEqual(rule, loaded_rule)


class ControlEdgeTest(unittest.TestCase):

    def test_control_edge(self):
        condition = MeetAllEventCondition()
        action = JobAction.START
        rule = SchedulingRule(condition, action)
        edge = ControlEdge('task', rule)
        self.assertEqual('*', edge.source)
        self.assertEqual('task', edge.destination)
        self.assertEqual(rule, edge.scheduling_rule)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""Dag sub-commands"""
import ast
import errno
import json
import logging
import signal
import subprocess
import sys

from graphviz.dot import Dot

from airflow import settings
from airflow.api.client import get_current_api_client
from airflow.cli.simple_table import AirflowConsole
from airflow.configuration import conf
from airflow.exceptions import AirflowException, BackfillUnfinished
from airflow.executors.debug_executor import DebugExecutor
from airflow.jobs.base_job import BaseJob
from airflow.models import DagBag, DagModel, DagRun, TaskInstance
from airflow.models.dag import DAG
from airflow.utils import cli as cli_utils
from airflow.utils.cli import (
    get_dag,
    get_dag_by_file_location,
    process_subdir,
    sigint_handler,
    suppress_logs_and_warning,
)
from airflow.utils.dot_renderer import render_dag
from airflow.utils.session import create_session, provide_session
from airflow.utils.state import State


@cli_utils.action_logging
def dag_backfill(args, dag=None):
    """Creates backfill job or dry run for a DAG"""
    logging.basicConfig(level=settings.LOGGING_LEVEL, format=settings.SIMPLE_LOG_FORMAT)

    signal.signal(signal.SIGTERM, sigint_handler)

    import warnings

    warnings.warn(
        '--ignore-first-depends-on-past is deprecated as the value is always set to True',
        category=PendingDeprecationWarning,
    )

    if args.ignore_first_depends_on_past is False:
        args.ignore_first_depends_on_past = True

    dag = dag or get_dag(args.subdir, args.dag_id)

    if not args.start_date and not args.end_date:
        raise AirflowException("Provide a start_date and/or end_date")

    # If only one date is passed, using same as start and end
    args.end_date = args.end_date or args.start_date
    args.start_date = args.start_date or args.end_date

    if args.task_regex:
        dag = dag.partial_subset(
            task_ids_or_regex=args.task_regex, include_upstream=not args.ignore_dependencies
        )

    run_conf = None
    if args.conf:
        run_conf = json.loads(args.conf)

    if args.dry_run:
        print(f"Dry run of DAG {args.dag_id} on {args.start_date}")
        for task in dag.tasks:
            print(f"Task {task.task_id}")
            ti = TaskInstance(task, args.start_date)
            ti.dry_run()
    else:
        if args.reset_dagruns:
            DAG.clear_dags(
                [dag],
                start_date=args.start_date,
                end_date=args.end_date,
                confirm_prompt=not args.yes,
                include_subdags=True,
                dag_run_state=State.NONE,
            )

        dag.run(
            start_date=args.start_date,
            end_date=args.end_date,
            mark_success=args.mark_success,
            local=args.local,
            donot_pickle=(args.donot_pickle or conf.getboolean('core', 'donot_pickle')),
            ignore_first_depends_on_past=args.ignore_first_depends_on_past,
            ignore_task_deps=args.ignore_dependencies,
            pool=args.pool,
            delay_on_limit_secs=args.delay_on_limit,
            verbose=args.verbose,
            conf=run_conf,
            rerun_failed_tasks=args.rerun_failed_tasks,
            run_backwards=args.run_backwards,
        )


@cli_utils.action_logging
def dag_trigger(args):
    """Creates a dag run for the specified dag"""
    api_client = get_current_api_client()
    try:
        message = api_client.trigger_dag(
            dag_id=args.dag_id, run_id=args.run_id, conf=args.conf, execution_date=args.exec_date
        )
        print(message)
    except OSError as err:
        raise AirflowException(err)


@cli_utils.action_logging
def dag_delete(args):
    """Deletes all DB records related to the specified dag"""
    api_client = get_current_api_client()
    if (
        args.yes
        or input("This will drop all existing records related to the specified DAG. Proceed? (y/n)").upper()
        == "Y"
    ):
        try:
            message = api_client.delete_dag(dag_id=args.dag_id)
            print(message)
        except OSError as err:
            raise AirflowException(err)
    else:
        print("Cancelled")


@cli_utils.action_logging
def dag_pause(args):
    """Pauses a DAG"""
    set_is_paused(True, args)


@cli_utils.action_logging
def dag_unpause(args):
    """Unpauses a DAG"""
    set_is_paused(False, args)


def set_is_paused(is_paused, args):
    """Sets is_paused for DAG by a given dag_id"""
    DagModel.get_dagmodel(args.dag_id).set_is_paused(
        is_paused=is_paused,
    )

    print("Dag: {}, paused: {}".format(args.dag_id, str(is_paused)))


def dag_show(args):
    """Displays DAG or saves it's graphic representation to the file"""
    dag = get_dag(args.subdir, args.dag_id)
    dot = render_dag(dag)
    filename = args.save
    imgcat = args.imgcat

    if filename and imgcat:
        raise SystemExit(
            "Option --save and --imgcat are mutually exclusive. "
            "Please remove one option to execute the command.",
        )
    elif filename:
        _save_dot_to_file(dot, filename)
    elif imgcat:
        _display_dot_via_imgcat(dot)
    else:
        print(dot.source)


def _display_dot_via_imgcat(dot: Dot):
    data = dot.pipe(format='png')
    try:
        proc = subprocess.Popen("imgcat", stdout=subprocess.PIPE, stdin=subprocess.PIPE)
    except OSError as e:
        if e.errno == errno.ENOENT:
            raise SystemExit(
                "Failed to execute. Make sure the imgcat executables are on your systems \'PATH\'"
            )
        else:
            raise
    out, err = proc.communicate(data)
    if out:
        print(out.decode('utf-8'))
    if err:
        print(err.decode('utf-8'))


def _save_dot_to_file(dot: Dot, filename: str):
    filename_without_ext, _, ext = filename.rpartition('.')
    dot.render(filename=filename_without_ext, format=ext, cleanup=True)
    print(f"File {filename} saved")


@cli_utils.action_logging
def dag_state(args):
    """
    Returns the state (and conf if exists) of a DagRun at the command line.
    >>> airflow dags state tutorial 2015-01-01T00:00:00.000000
    running
    >>> airflow dags state a_dag_with_conf_passed 2015-01-01T00:00:00.000000
    failed, {"name": "bob", "age": "42"}
    """
    if args.subdir:
        dag = get_dag(args.subdir, args.dag_id)
    else:
        dag = get_dag_by_file_location(args.dag_id)
    dr = DagRun.find(dag.dag_id, execution_date=args.execution_date)
    out = dr[0].state if dr else None
    conf_out = ''
    if out and dr[0].conf:
        conf_out = ', ' + json.dumps(dr[0].conf)
    print(str(out) + conf_out)


@cli_utils.action_logging
def dag_next_execution(args):
    """
    Returns the next execution datetime of a DAG at the command line.
    >>> airflow dags next-execution tutorial
    2018-08-31 10:38:00
    """
    dag = get_dag(args.subdir, args.dag_id)

    if dag.get_is_paused():
        print("[INFO] Please be reminded this DAG is PAUSED now.", file=sys.stderr)

    latest_execution_date = dag.get_latest_execution_date()
    if latest_execution_date:
        next_execution_dttm = dag.following_schedule(latest_execution_date)

        if next_execution_dttm is None:
            print(
                "[WARN] No following schedule can be found. "
                + "This DAG may have schedule interval '@once' or `None`.",
                file=sys.stderr,
            )
            print(None)
        else:
            print(next_execution_dttm)

            for _ in range(1, args.num_executions):
                next_execution_dttm = dag.following_schedule(next_execution_dttm)
                print(next_execution_dttm)
    else:
        print("[WARN] Only applicable when there is execution record found for the DAG.", file=sys.stderr)
        print(None)


@cli_utils.action_logging
@suppress_logs_and_warning()
def dag_list_dags(args):
    """Displays dags with or without stats at the command line"""
    dagbag = DagBag(process_subdir(args.subdir))
    AirflowConsole().print_as(
        data=sorted(dagbag.dags.values(), key=lambda d: d.dag_id),
        output=args.output,
        mapper=lambda x: {
            "dag_id": x.dag_id,
            "filepath": x.filepath,
            "owner": x.owner,
            "paused": x.get_is_paused(),
        },
    )


@cli_utils.action_logging
@suppress_logs_and_warning()
def dag_report(args):
    """Displays dagbag stats at the command line"""
    dagbag = DagBag(process_subdir(args.subdir))
    AirflowConsole().print_as(
        data=dagbag.dagbag_stats,
        output=args.output,
        mapper=lambda x: {
            "file": x.file,
            "duration": x.duration,
            "dag_num": x.dag_num,
            "task_num": x.task_num,
            "dags": sorted(ast.literal_eval(x.dags)),
        },
    )


@cli_utils.action_logging
@suppress_logs_and_warning()
def dag_list_jobs(args, dag=None):
    """Lists latest n jobs"""
    queries = []
    if dag:
        args.dag_id = dag.dag_id
    if args.dag_id:
        dagbag = DagBag()

        if args.dag_id not in dagbag.dags:
            error_message = f"Dag id {args.dag_id} not found"
            raise AirflowException(error_message)
        queries.append(BaseJob.dag_id == args.dag_id)

    if args.state:
        queries.append(BaseJob.state == args.state)

    fields = ['dag_id', 'state', 'job_type', 'start_date', 'end_date']
    with create_session() as session:
        all_jobs = (
            session.query(BaseJob)
            .filter(*queries)
            .order_by(BaseJob.start_date.desc())
            .limit(args.limit)
            .all()
        )
        all_jobs = [{f: str(job.__getattribute__(f)) for f in fields} for job in all_jobs]

    AirflowConsole().print_as(
        data=all_jobs,
        output=args.output,
    )


@cli_utils.action_logging
@suppress_logs_and_warning()
def dag_list_dag_runs(args, dag=None):
    """Lists dag runs for a given DAG"""
    if dag:
        args.dag_id = dag.dag_id

    dagbag = DagBag()

    if args.dag_id is not None and args.dag_id not in dagbag.dags:
        error_message = f"Dag id {args.dag_id} not found"
        raise AirflowException(error_message)

    state = args.state.lower() if args.state else None
    dag_runs = DagRun.find(
        dag_id=args.dag_id,
        state=state,
        no_backfills=args.no_backfill,
        execution_start_date=args.start_date,
        execution_end_date=args.end_date,
    )

    dag_runs.sort(key=lambda x: x.execution_date, reverse=True)
    AirflowConsole().print_as(
        data=dag_runs,
        output=args.output,
        mapper=lambda dr: {
            "dag_id": dr.dag_id,
            "run_id": dr.run_id,
            "state": dr.state,
            "execution_date": dr.execution_date.isoformat(),
            "start_date": dr.start_date.isoformat() if dr.start_date else '',
            "end_date": dr.end_date.isoformat() if dr.end_date else '',
        },
    )


@provide_session
@cli_utils.action_logging
def dag_test(args, session=None):
    """Execute one single DagRun for a given DAG and execution date, using the DebugExecutor."""
    dag = get_dag(subdir=args.subdir, dag_id=args.dag_id)
    dag.clear(start_date=args.execution_date, end_date=args.execution_date, dag_run_state=State.NONE)
    try:
        dag.run(executor=DebugExecutor(), start_date=args.execution_date, end_date=args.execution_date)
    except BackfillUnfinished as e:
        print(str(e))

    show_dagrun = args.show_dagrun
    imgcat = args.imgcat_dagrun
    filename = args.save_dagrun
    if show_dagrun or imgcat or filename:
        tis = (
            session.query(TaskInstance)
            .filter(
                TaskInstance.dag_id == args.dag_id,
                TaskInstance.execution_date == args.execution_date,
            )
            .all()
        )

        dot_graph = render_dag(dag, tis=tis)
        print()
        if filename:
            _save_dot_to_file(dot_graph, filename)
        if imgcat:
            _display_dot_via_imgcat(dot_graph)
        if show_dagrun:
            print(dot_graph.source)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
import os
import time
from unittest import mock

from ai_flow.common.properties import Properties
from ai_flow.common.status import Status
from ai_flow.meta.dataset_meta import DataType, DatasetMeta, Schema
from ai_flow.meta.job_meta import State
from ai_flow.meta.metric_meta import MetricType
from ai_flow.model_center.entity.registered_model_detail import RegisteredModelDetail
from ai_flow.protobuf.message_pb2 import RESOURCE_ALREADY_EXISTS, \
    INVALID_PARAMETER_VALUE
from ai_flow.endpoint.server.exception import AIFlowException
from ai_flow.test.endpoint import random_str


class AbstractTestStore(object):

    """test dataset"""

    def test_save_dataset_get_dataset_by_id_and_name(self):
        response = self.store.register_dataset(name='dataset', data_format='csv',
                                               properties=Properties({'a': 'b'}),
                                               name_list=['a'], type_list=[DataType.STRING])
        self.assertEqual(response.name, 'dataset')
        response_id = self.store.get_dataset_by_id(response.uuid)
        self.assertEqual('dataset', response_id.name)
        response_by_name = self.store.get_dataset_by_name('dataset')
        self.assertEqual(response_by_name.name, 'dataset')

    def test_save_dataset_with_catalog_by_id_and_name(self):
        response = self.store.register_dataset_with_catalog(name='dataset',
                                                            catalog_name='catalog', catalog_type='kafka',
                                                            catalog_database='my_db',
                                                            catalog_table='my_table', catalog_connection_uri='/path')
        print(response.to_json_dict())
        response_id = self.store.get_dataset_by_id(response.uuid)
        self.assertEqual('dataset', response_id.name)
        response_by_name = self.store.get_dataset_by_name('dataset')
        self.assertEqual(response_by_name.name, 'dataset')
        print(response_by_name.to_json_dict(), response_id.to_json_dict())

    def test_double_register_dataset(self):
        dataset_1 = self.store.register_dataset(name='dataset', data_format='csv', properties=Properties({'a': 'b'}),
                                                name_list=['a'], type_list=[DataType.STRING])
        dataset_2 = self.store.register_dataset(name='dataset', data_format='csv', properties=Properties({'a': 'b'}),
                                                name_list=['a'], type_list=[DataType.STRING])
        self.assertEqual(dataset_1.uuid, dataset_2.uuid)
        self.assertEqual(dataset_1.schema.to_json_dict(), dataset_2.schema.to_json_dict())
        self.assertEqual(dataset_1.schema.to_json_dict(), dataset_2.schema.to_json_dict())
        self.assertRaises(AIFlowException, self.store.register_dataset, name='dataset',
                          data_format='txt',
                          properties=Properties({'a': 'b'}),
                          name_list=['a'], type_list=[DataType.STRING])

    def test_double_register_dataset_with_catalog(self):
        dataset_1 = self.store.register_dataset_with_catalog(name='dataset',
                                                             catalog_name='catalog', catalog_type='kafka',
                                                             catalog_database='my_db',
                                                             catalog_table='my_table', catalog_connection_uri='/path')
        dataset_2 = self.store.register_dataset_with_catalog(name='dataset',
                                                             catalog_name='catalog', catalog_type='kafka',
                                                             catalog_database='my_db',
                                                             catalog_table='my_table', catalog_connection_uri='/path')
        self.assertEqual(dataset_1.uuid, dataset_2.uuid)
        self.assertEqual(dataset_1.schema.to_json_dict(), dataset_2.schema.to_json_dict())
        self.assertEqual(dataset_1.schema.to_json_dict(), dataset_2.schema.to_json_dict())
        self.assertRaises(AIFlowException, self.store.register_dataset, name='dataset',
                          data_format='csv', properties=Properties({'a': 'b'}),
                          name_list=['a'], type_list=[DataType.STRING])

    def test_list_datasets(self):
        self.store.register_dataset(name='dataset_1', data_format='csv', description='it is mq data',
                                    uri='mysql://', properties=Properties({'a': 'b'}), name_list=['a'],
                                    type_list=[DataType.INT32])
        self.store.register_dataset(name='dataset_2', data_format='npz', description='it is',
                                    uri='mysql://', properties=Properties({'a': 'b'}), name_list=['a'],
                                    type_list=[DataType.INT32])
        response_list = self.store.list_datasets(5, 0)
        self.assertEqual(len(response_list), 2)
        self.assertEqual('dataset_1', response_list[0].name)
        self.assertEqual('dataset_2', response_list[1].name)

    def test_save_datasets_list_datasets(self):
        schema = Schema(name_list=['a'],
                        type_list=[DataType.STRING])
        dataset_1 = DatasetMeta(name='dataset1', data_format='csv',
                                properties=Properties({'a': 'b'}), schema=schema)
        dataset_2 = DatasetMeta(name='dataset2')
        response = self.store.register_datasets([dataset_1, dataset_2])
        self.assertEqual(len(response), 2)
        self.assertEqual(1, response[0].uuid)
        self.assertEqual(2, response[1].uuid)
        response_list = self.store.list_datasets(2, 0)
        self.assertEqual(2, len(response_list))
        self.assertEqual('dataset1', response_list[0].name)
        self.assertEqual('dataset2', response_list[1].name)

    def test_delete_dataset(self):
        self.store.register_dataset(name='dataset', data_format='csv')
        self.assertEqual(Status.OK, self.store.delete_dataset_by_id(1))
        self.assertIsNone(self.store.get_dataset_by_name(dataset_name='dataset'))
        self.store.register_dataset(name='dataset', data_format='csv')
        self.assertEqual(Status.OK, self.store.delete_dataset_by_id(2))
        self.assertIsNone(self.store.get_dataset_by_name(dataset_name='dataset'))
        self.store.register_dataset(name='dataset', data_format='csv')
        self.assertEqual(Status.OK, self.store.delete_dataset_by_name('dataset'))
        self.assertIsNone(self.store.get_dataset_by_name(dataset_name='dataset'))
        self.store.register_dataset(name='another_dataset',
                                    data_format='csv')
        self.assertEqual(Status.OK, self.store.delete_dataset_by_name('another_dataset'))
        self.store.register_dataset(name='another_dataset',
                                    data_format='csv')
        self.assertEqual(Status.OK, self.store.delete_dataset_by_name('another_dataset'))
        self.assertIsNone(self.store.get_dataset_by_name(dataset_name='another_dataset'))

    def test_update_dataset(self):
        self.store.register_dataset(name='dataset', data_format='csv')
        update_dataset = self.store.update_dataset(dataset_name='dataset',
                                                   data_format='json',
                                                   description='it is a training dataset',
                                                   properties=Properties({'title': 'iris_training'}),
                                                   name_list=['a'], type_list=[DataType.FLOAT32])
        self.store.register_dataset_with_catalog(name='dataset_withcatalog',
                                                 catalog_name='my_hive', catalog_database='default',
                                                 catalog_connection_uri='/path/to/conf', catalog_type='hive',
                                                 catalog_table='my_table')
        update_dataset_1 = self.store.update_dataset(dataset_name='dataset_withcatalog',
                                                     catalog_name='my_hive', catalog_database='my_db',
                                                     catalog_connection_uri='/path/to/conf', catalog_type='hive')
        self.assertEqual(update_dataset.schema.name_list, ['a'])
        self.assertEqual(update_dataset_1.catalog_database, 'my_db')

    """test workflow"""
    def test_save_workflow_get_workflow_by_id_and_name(self):
        project_response = self.store.register_project(name='project', uri='www.code.com')
        self.assertEqual(project_response.uuid, 1)
        response = self.store.register_workflow(name='workflow',
                                                project_id=project_response.uuid,
                                                properties=Properties({'a': 'b'}))
        self.assertEqual(response.uuid, 1)
        self.assertEqual(response.properties, Properties({'a': 'b'}))
        response_by_id = self.store.get_workflow_by_id(response.uuid)
        response_by_name = self.store.get_workflow_by_name(project_response.name, response.name)
        self.assertEqual('workflow', response_by_id.name)
        self.assertEqual('workflow', response_by_name.name)
        self.assertEqual(Properties({'a': 'b'}), response_by_id.properties)
        self.assertEqual(Properties({'a': 'b'}), response_by_name.properties)

    def test_double_register_workflow(self):
        project_response = self.store.register_project(name='project', uri='www.code.com')
        project_response2 = self.store.register_project(name='project2', uri='www.code.com')
        self.store.register_workflow(name='workflow', project_id=project_response.uuid)
        self.store.register_workflow(name='workflow', project_id=project_response2.uuid)
        self.assertRaises(AIFlowException, self.store.register_workflow, name='workflow',
                          project_id=project_response.uuid)

    def test_list_workflows(self):
        project_response = self.store.register_project(name='project', uri='www.code.com')
        self.store.register_workflow(name='workflow1', project_id=project_response.uuid)
        self.store.register_workflow(name='workflow2', project_id=project_response.uuid)
        response_list = self.store.list_workflows(project_response.name, 2, 0)
        self.assertEqual('workflow1', response_list[0].name)
        self.assertEqual('workflow2', response_list[1].name)

    def test_delete_workflow(self):
        project_response = self.store.register_project(name='project', uri='www.code.com')
        response = self.store.register_workflow(name='workflow', project_id=project_response.uuid)
        self.assertEqual(Status.OK, self.store.delete_workflow_by_name(project_name=project_response.name,
                                                                       workflow_name='workflow'))
        self.assertIsNone(self.store.get_workflow_by_id(response.uuid))

        response = self.store.register_workflow(name='workflow', project_id=project_response.uuid)
        self.assertEqual(Status.OK, self.store.delete_workflow_by_id(response.uuid))
        self.assertIsNone(self.store.get_workflow_by_id(response.uuid))

    def test_update_workflow(self):
        project_response = self.store.register_project(name='project', uri='www.code.com')
        response = self.store.register_workflow(name='workflow',
                                                project_id=project_response.uuid,
                                                properties=Properties({'a': 'b'}))

        updated_workflow = self.store.update_workflow(project_name=project_response.name,
                                                      workflow_name='workflow',
                                                      properties=Properties({'a': 'c'}))
        self.assertEqual(updated_workflow.properties, Properties({'a': 'c'}))

    """test project"""

    def test_save_project_get_project_by_id_and_name(self):
        response = self.store.register_project(name='project', uri='www.code.com')
        self.assertEqual(response.uuid, 1)
        response_id = self.store.get_project_by_id(response.uuid)
        response_name = self.store.get_project_by_name('project')
        self.assertEqual('project', response_id.name)
        self.assertEqual('project', response_name.name)
        print(response_id)

    def test_double_register_project(self):
        self.store.register_project(name='project', uri='www.code.com')
        self.store.register_project(name='project', uri='www.code.com')
        self.assertRaises(AIFlowException, self.store.register_project, name='project',
                          uri='www.code2.com')

    def test_list_project(self):
        self.store.register_project(name='project', uri='www.code.com')
        self.store.register_project(name='project1', uri='www.code.com')
        response_list = self.store.list_project(2, 0)
        self.assertEqual(2, len(response_list))
        self.assertEqual('project', response_list[0].name)
        self.assertEqual('project1', response_list[1].name)
        print(response_list[1])

    def test_delete_project_by_id(self):
        self.store.register_project(name='project', uri='www.code.com')
        self.store.register_model_relation(name='model', project_id=1)
        self.store.register_model_version_relation(version='1', model_id=1,
                                                   project_snapshot_id=None)
        self.assertEqual(self.store.get_project_by_id(1).name, 'project')
        self.assertEqual(self.store.get_model_relation_by_id(1).name, 'model')
        self.assertEqual(self.store.get_model_version_relation_by_version('1', '1').version, '1')
        self.assertEqual(Status.OK, self.store.delete_project_by_id(1))
        self.assertIsNone(self.store.get_project_by_id(1))
        self.assertIsNone(self.store.get_model_relation_by_id(1))
        self.assertIsNone(self.store.get_model_version_relation_by_version('1', '1'))
        self.assertIsNone(self.store.list_project(1, 0))
        self.assertIsNone(self.store.list_model_relation(1, 0))
        self.assertIsNone(self.store.list_model_version_relation(1, 1, 0))
        self.store.register_project(name='project', uri='www.code.com')
        self.store.register_model_relation(name='model', project_id=2)
        self.store.register_model_version_relation(version='1', model_id=2,
                                                   project_snapshot_id=None)
        self.assertEqual(Status.OK, self.store.delete_project_by_id(2))

    def test_delete_project_by_name(self):
        self.store.register_project(name='project', uri='www.code.com')
        self.store.register_model_relation(name='model', project_id=1)
        self.store.register_model_version_relation(version='1', model_id=1,
                                                   project_snapshot_id=None)
        self.assertEqual(self.store.get_project_by_id(1).name, 'project')
        self.assertEqual(self.store.get_model_relation_by_id(1).name, 'model')
        self.assertEqual(self.store.get_model_version_relation_by_version('1', '1').version, '1')
        self.assertEqual(Status.OK, self.store.delete_project_by_name('project'))
        self.assertIsNone(self.store.get_project_by_id(1))
        self.assertIsNone(self.store.get_model_relation_by_id(1))
        self.assertIsNone(self.store.get_model_version_relation_by_version('1', '1'))
        self.assertIsNone(self.store.list_project(1, 0))
        self.assertIsNone(self.store.list_model_relation(1, 0))
        self.assertIsNone(self.store.list_model_version_relation(1, 1, 0))
        self.store.register_project(name='project', uri='www.code.com')
        self.store.register_model_relation(name='model', project_id=2)
        self.store.register_model_version_relation(version='1', model_id=2,
                                                   project_snapshot_id=None)
        self.assertEqual(Status.OK, self.store.delete_project_by_name('project'))

    def test_update_project(self):
        self.store.register_project(name='project', uri='www.code.com')
        update_project = self.store.update_project(project_name='project', uri='git@alibaba')
        self.assertEqual(update_project.uri, 'git@alibaba')
        self.assertIsNone(update_project.properties)

    """test model """

    def test_save_model_get_id_and_name(self):
        self.store.register_project(name='project', uri='www.code.com')
        response = self.store.register_model_relation(name='model', project_id=1)
        self.assertEqual(response.name, 'model')
        self.assertEqual(self.store.get_model_relation_by_id(response.uuid).name, 'model')
        self.assertEqual(self.store.get_model_relation_by_name('model').name, 'model')
        print(self.store.get_model_relation_by_id(response.uuid))

    def test_list_model(self):
        self.store.register_project(name='project', uri='www.code.com')
        self.store.register_model_relation(name='model', project_id=1)
        self.store.register_model_relation(name='model1', project_id=1)
        self.assertEqual(2, len(self.store.list_model_relation(2, 0)))
        self.assertEqual('model', self.store.list_model_relation(2, 0)[0].name)
        self.assertEqual('model1', self.store.list_model_relation(2, 0)[1].name)

    def test_delete_model_by_id(self):
        self.store.register_project(name='project', uri='www.code.com')
        self.store.register_model_relation(name='model', project_id=1)
        response = self.store.register_model_version_relation(version='1', model_id=1, project_snapshot_id=None)
        self.assertEqual(response.version, '1')
        self.assertEqual(self.store.get_model_version_relation_by_version('1', 1).version, '1')
        self.assertEqual(self.store.get_model_relation_by_name('model').name, 'model')
        self.assertEqual(Status.OK, self.store.delete_model_relation_by_id(1))
        self.assertIsNone(self.store.get_model_version_relation_by_version('1', '1'))
        self.assertIsNone(self.store.get_model_relation_by_name('model'))
        self.store.register_model_relation(name='model', project_id=1)
        self.store.register_model_version_relation(version='1', model_id=2, project_snapshot_id=None)
        self.assertEqual(Status.OK, self.store.delete_model_relation_by_id(2))

    def test_delete_model_by_name(self):
        self.store.register_project(name='project', uri='www.code.com')
        self.store.register_model_relation(name='model', project_id=1)
        response = self.store.register_model_version_relation(version='1', model_id=1, project_snapshot_id=None)
        self.assertEqual(response.version, '1')
        self.assertEqual(self.store.get_model_version_relation_by_version('1', '1').version, '1')
        self.assertEqual(self.store.get_model_relation_by_name('model').name, 'model')
        self.assertEqual(Status.OK, self.store.delete_model_relation_by_name('model'))
        self.assertIsNone(self.store.get_model_version_relation_by_version('1', '1'))
        self.assertIsNone(self.store.get_model_relation_by_name('model'))
        self.store.register_model_relation(name='model', project_id=1)
        self.store.register_model_version_relation(version='1', model_id=2, project_snapshot_id=None)
        self.assertEqual(Status.OK, self.store.delete_model_relation_by_name('model'))

    def test_double_register_model_relation(self):
        self.store.register_project(name='project', uri='www.code.com')
        self.store.register_model_relation(name='model', project_id=1)
        self.store.register_model_relation(name='model', project_id=1)
        self.assertRaises(AIFlowException, self.store.register_model_relation, name='model', project_id=2)

    """test model version"""

    def test_save_model_version_get_by_version(self):
        self.store.register_project(name='project', uri='www.code.com')
        self.store.register_model_relation(name='model', project_id=1)
        response = self.store.register_model_version_relation(version='1', model_id=1, project_snapshot_id=None)
        self.assertEqual(response.version, '1')
        self.assertEqual(self.store.get_model_version_relation_by_version(version_name='1', model_id=1).version, '1')

    def test_list_model_version(self):
        self.store.register_project(name='project', uri='www.code.com')
        self.store.register_model_relation(name='model', project_id=1)
        self.store.register_model_version_relation(version='1', model_id=1, project_snapshot_id=None)
        self.store.register_model_version_relation(version='2', model_id=1, project_snapshot_id=None)
        self.assertEqual(len(self.store.list_model_version_relation(1, 2, 0)), 2)
        self.assertEqual(self.store.list_model_version_relation(1, 2, 0)[0].version, '1')
        self.assertEqual(self.store.list_model_version_relation(1, 2, 0)[1].version, '2')

    def test_delete_model_version_by_version(self):
        self.store.register_project(name='project', uri='www.code.com')
        self.store.register_model_relation(name='model', project_id=1)
        self.store.register_model_version_relation(version='1', model_id=1, project_snapshot_id=None)
        self.assertEqual(self.store.get_model_version_relation_by_version('1', 1).version, '1')
        self.assertEqual(Status.OK, self.store.delete_model_version_relation_by_version('1', 1))
        self.assertIsNone(self.store.get_model_version_relation_by_version('1', 1))
        self.store.register_model_version_relation(version='1', model_id=1, project_snapshot_id=None)
        self.assertEqual(Status.OK, self.store.delete_model_version_relation_by_version('1', 1))

    """test artifact"""

    def test_register_artifact_get_and_list_artifact(self):
        artifact = self.store.register_artifact(name='artifact_result', uri='../..')
        self.assertEqual(artifact.uuid, self.store.get_artifact_by_id(artifact.uuid).uuid)
        self.assertEqual(artifact.uri, self.store.get_artifact_by_name('artifact_result').uri)
        self.store.register_artifact(name='artifact_result_1', uri='../..')
        self.assertEqual(2, len(self.store.list_artifact(2, 0)))
        self.assertEqual(Status.OK, self.store.delete_artifact_by_id(1))
        self.assertEqual(Status.OK, self.store.delete_artifact_by_name('artifact_result_1'))
        self.assertEqual(Status.ERROR, self.store.delete_artifact_by_name('no artifact'))
        self.assertIsNone(self.store.get_artifact_by_id(1))
        self.assertIsNone(self.store.get_artifact_by_name('artifact_result_1'))
        print(artifact.to_json_dict())

    def test_double_register_artifact(self):
        artifact_1 = self.store.register_artifact(name='artifact_result', uri='../..')
        artifact_2 = self.store.register_artifact(name='artifact_result', uri='../..')
        self.assertEqual(artifact_1.to_json_dict(), artifact_2.to_json_dict())
        self.assertRaises(AIFlowException, self.store.register_artifact, name='artifact_result', uri='.')

    def test_update_artifact(self):
        artifact = self.store.register_artifact(name='artifact_result', uri='../..')
        update_artifact = self.store.update_artifact(name='artifact_result')
        self.assertEqual(update_artifact.uri, artifact.uri)
        self.assertIsNone(update_artifact.properties)
        self.assertIsNone(update_artifact.artifact_type)

    def _create_registered_model(self, model_name, model_desc='model desc'):
        return self.store.create_registered_model(model_name, model_desc)

    def _create_model_version(self, model_name, model_path='path/to/source',
                              model_type='{"flavor.version":1}', version_desc='model version desc'):
        return self.store.create_model_version(model_name, model_path, model_type, version_desc)

    def test_create_registered_model(self):
        model_name1 = random_str() + 'ABcd'
        model_desc1 = 'test_create_registered_model1'
        register_model1 = self._create_registered_model(model_name1, model_desc1)
        self.assertEqual(register_model1.model_name, model_name1)

        # error on duplicate
        model_desc2 = 'test_create_registered_model2'
        with self.assertRaises(AIFlowException) as exception_context:
            self._create_registered_model(model_name1, model_desc2)
        assert exception_context.exception.error_code == RESOURCE_ALREADY_EXISTS

        # slightly different name is ok
        for model_name2 in [model_name1 + 'extra', model_name1 + model_name1]:
            register_model2 = self._create_registered_model(model_name2, model_desc2)
            self.assertEqual(register_model2.model_name, model_name2)

    def test_double_create_model(self):
        model_name1 = random_str() + 'ABcd'
        model_desc1 = 'test_create_registered_model1'
        self._create_registered_model(model_name1, model_desc1)
        model_name2 = random_str() + 'ABcd'
        model_desc2 = 'test_create_registered_model1'
        self._create_registered_model(model_name2, model_desc2)
        self.assertRaises(AIFlowException, self._create_registered_model, model_name2, model_desc=' ')

    def test_update_registered_model(self):
        model_name1 = random_str() + 'ABcd'
        model_desc1 = 'test_update_registered_model'
        register_model1 = self._create_registered_model(model_name1, model_desc1)
        register_model_detail1 = self.store.get_registered_model_detail(register_model1)
        self.assertEqual(register_model1.model_name, model_name1)

        # update model name
        register_model2 = self.store.update_registered_model(register_model1,
                                                             model_name='NewName')
        register_model_detail2 = self.store.get_registered_model_detail(register_model2)
        self.assertEqual(register_model2.model_name, 'NewName')
        self.assertEqual(register_model_detail2.model_name, 'NewName')

        # update model description
        register_model4 = self.store.update_registered_model(register_model2,
                                                             model_desc='update_registered_model_desc')
        register_model_detail4 = self.store.get_registered_model_detail(register_model4)
        self.assertEqual(register_model4.model_name, 'NewName')
        self.assertEqual(register_model_detail4.model_name, 'NewName')
        self.assertEqual(register_model_detail4.model_desc, 'update_registered_model_desc')

        # update both model name and description
        register_model5 = self.store.update_registered_model(register_model4,
                                                             model_name='AnotherName',
                                                             model_desc='TEST')
        register_model_detail5 = self.store.get_registered_model_detail(register_model5)
        self.assertEqual(register_model5.model_name, 'AnotherName')
        self.assertEqual(register_model_detail5.model_name, 'AnotherName')
        self.assertEqual(register_model_detail5.model_desc, 'TEST')

        # new models with old names
        self._create_registered_model(model_name1, model_desc1)
        register_model5 = self._create_registered_model('NewName', 'update_registered_model_desc')

        # cannot rename model to conflict with an existing model
        with self.assertRaises(AIFlowException) as exception_context:
            self.store.update_registered_model(register_model5, 'AnotherName')
        assert exception_context.exception.error_code == RESOURCE_ALREADY_EXISTS

    def test_delete_registered_model(self):
        registered_model = self._create_registered_model('model_for_delete_RM')
        register_model_detail = self.store.get_registered_model_detail(registered_model)
        self.assertEqual(register_model_detail.model_name, 'model_for_delete_RM')

        # delete model
        self.store.delete_registered_model(registered_model)

    def test_list_registered_model(self):
        self._create_registered_model('M')
        registered_models = self.store.list_registered_models()
        self.assertEqual(len(registered_models), 1)
        self.assertEqual(registered_models[0].model_name, 'M')
        self.assertIsInstance(registered_models[0], RegisteredModelDetail)

        self._create_registered_model('N')
        self.assertEqual(set([registered_model.model_name for registered_model in self.store.list_registered_models()]),
                         {'M', 'N'})

        self._create_registered_model('NN')
        self._create_registered_model('NM')
        self._create_registered_model('MN')
        self._create_registered_model('NNO')
        self.assertEqual(set([rm.model_name for rm in self.store.list_registered_models()]),
                         {'M', 'N', 'NN', 'NM', 'MN', 'NNO'})

    def test_get_registered_model_detail(self):
        model_name = 'test_model'
        model_desc = 'test_get_registered_model_detail'
        # use fake clock
        with mock.patch('time.time') as mock_time:
            mock_time.return_value = 1234
            registered_model = self._create_registered_model(model_name, model_desc)
            self.assertEqual(registered_model.model_name, model_name)
        register_model_detail = self.store.get_registered_model_detail(registered_model)
        self.assertEqual(register_model_detail.model_name, model_name)
        self.assertEqual(register_model_detail.model_desc, model_desc)
        self.assertEqual(register_model_detail.latest_model_version, None)

    def test_create_model_version(self):
        model_name = 'test_for_create_model_version'
        self._create_registered_model(model_name)
        with mock.patch('time.time') as mock_time:
            mock_time.return_value = 456778
            model_version1 = self._create_model_version(model_name)
            self.assertEqual(model_version1.model_name, model_name)
            self.assertEqual(model_version1.model_version, '1')

        model_version_detail1 = self.store.get_model_version_detail(model_version1)
        self.assertEqual(model_version_detail1.model_name, model_name)
        self.assertEqual(model_version_detail1.model_version, '1')
        self.assertEqual(model_version_detail1.model_path, 'path/to/source')
        self.assertEqual(model_version_detail1.version_desc, 'model version desc')
        self.assertEqual(model_version_detail1.version_status, 'READY')
        self.assertEqual(model_version_detail1.current_stage, 'Generated')
        self.store.update_model_version(model_version1, current_stage='Validated')
        self.assertEqual(self.store.get_latest_validated_model_version(model_name).model_version, '1')
        self.store.update_model_version(model_version1, current_stage='Deployed')
        self.assertEqual(self.store.get_deployed_model_version(model_name).model_version, '1')

        # new model versions for same name autoincrement versions
        model_version2 = self._create_model_version(model_name)
        model_version_detail2 = self.store.get_model_version_detail(model_version2)
        self.assertEqual(model_version2.model_version, '2')
        self.assertEqual(model_version_detail2.model_version, '2')

        model_version3 = self._create_model_version(model_name)
        model_version_detail3 = self.store.get_model_version_detail(model_version3)
        self.assertEqual(model_version3.model_version, '3')
        self.assertEqual(model_version_detail3.model_version, '3')

    def test_get_deployed_model_version(self):
        model_name = 'test_for_create_model_version'
        self._create_registered_model(model_name)
        model_version = self._create_model_version(model_name)
        serving_model_version = self.store.get_deployed_model_version(model_name)
        self.assertIsNone(serving_model_version)

        self.store.update_model_version(model_version=model_version,
                                        current_stage='DEPLOYED')
        serving_model_version = self.store.get_deployed_model_version(model_name)
        self.assertEqual(serving_model_version.model_version, model_version.model_version)
        self.assertRaises(AIFlowException, self.store.update_model_version,
                          model_version=model_version, current_stage='DEPLOYED')

    def test_get_latest_validated_model_version(self):
        model_name = 'test_for_get_validated_model_version'
        self._create_registered_model(model_name)
        model_version = self._create_model_version(model_name)
        serving_model_version = self.store.get_latest_validated_model_version(model_name)
        self.assertIsNone(serving_model_version)

        self.store.update_model_version(model_version=model_version,
                                        current_stage='VALIDATED')

        for i in range(10):
            model_version = self._create_model_version(model_name)
            self.store.update_model_version(model_version=model_version, current_stage='VALIDATED')
        serving_model_version = self.store.get_latest_validated_model_version(model_name)
        self.assertEqual(model_version.model_version, serving_model_version.model_version)

    def test_get_latest_generated_model_version(self):
        model_name = 'test_for_get_generated_model_version'
        self._create_registered_model(model_name)

        for i in range(10):
            model_version = self._create_model_version(model_name)
        generated_model_version = self.store.get_latest_generated_model_version(model_name)
        self.assertEqual(model_version.model_version, generated_model_version.model_version)

    def test_update_model_version(self):
        model_name = 'test_for_update_model_version'
        self._create_registered_model(model_name)
        model_version1 = self._create_model_version(model_name)
        model_version_detail1 = self.store.get_model_version_detail(model_version1)
        self.assertEqual(model_version_detail1.model_name, model_name)
        self.assertEqual(model_version_detail1.model_version, '1')
        self.assertEqual(model_version_detail1.current_stage, 'Generated')

        # update current stage
        self.store.update_model_version(model_version1, current_stage='Generated')
        model_version_detail2 = self.store.get_model_version_detail(model_version1)
        self.assertEqual(model_version_detail2.model_name, model_name)
        self.assertEqual(model_version_detail2.model_version, '1')
        self.assertEqual(model_version_detail2.current_stage, 'Generated')
        self.assertEqual(model_version_detail2.version_desc, 'model version desc')

        # update version description
        self.store.update_model_version(model_version1, version_desc='test model version')
        model_version_detail3 = self.store.get_model_version_detail(model_version1)
        self.assertEqual(model_version_detail3.model_name, model_name)
        self.assertEqual(model_version_detail3.model_version, '1')
        self.assertEqual(model_version_detail3.current_stage, 'Generated')
        self.assertEqual(model_version_detail3.version_desc, 'test model version')

        # update current stage and description
        self.store.update_model_version(model_version1, current_stage='Validated', version_desc='test version desc')
        model_version_detail4 = self.store.get_model_version_detail(model_version1)
        self.assertEqual(model_version_detail4.model_name, model_name)
        self.assertEqual(model_version_detail4.model_version, '1')
        self.assertEqual(model_version_detail4.current_stage, 'Validated')
        self.assertEqual(model_version_detail4.version_desc, 'test version desc')

        # only valid stages can be set
        with self.assertRaises(AIFlowException) as exception_context:
            self.store.update_model_version(model_version1, current_stage='unknown')
        assert exception_context.exception.error_code == INVALID_PARAMETER_VALUE

        # stages are case-insensitive and auto-corrected to system stage names
        for stage_name in ['DEPLOYED', 'deployed', 'DePloyEd']:
            self.store.update_model_version(model_version1, current_stage=stage_name)
            model_version_detail5 = self.store.get_model_version_detail(model_version1)
            self.assertEqual(model_version_detail5.current_stage, 'Deployed')

    def test_delete_model_version(self):
        model_name = 'test_delete_model_version'
        self._create_registered_model(model_name)
        model_version = self._create_model_version(model_name)
        model_version_detail = self.store.get_model_version_detail(model_version)
        self.assertEqual(model_version_detail.model_name, model_name)

        self.store.delete_model_version(model_version)

    def test_metric_meta(self):
        start_time = round(time.time())
        end_time = start_time + 1
        metric_meta = self.store.register_metric_meta(metric_name='test_metric_meta_1',
                                                      metric_type=MetricType.DATASET,
                                                      metric_desc='test dataset metric meta',
                                                      project_name='test_metric_meta_project_1',
                                                      dataset_name='test_metric_meta_dataset_1',
                                                      start_time=start_time, end_time=end_time,
                                                      uri='/tmp/metric', tags='test_metric_meta_tag',
                                                      properties={'a': 'a'})
        metric_meta = self.store.get_metric_meta(metric_meta.metric_name)
        self.assertEqual('test_metric_meta_1', metric_meta.metric_name)
        self.assertEqual(MetricType.DATASET, MetricType.value_of(metric_meta.metric_type))
        self.assertEqual('test dataset metric meta', metric_meta.metric_desc)
        self.assertEqual('test_metric_meta_project_1', metric_meta.project_name)
        self.assertEqual('test_metric_meta_dataset_1', metric_meta.dataset_name)
        self.assertEqual(start_time, metric_meta.start_time)
        self.assertEqual(end_time, metric_meta.end_time)
        self.assertEqual('/tmp/metric', metric_meta.uri)
        self.assertEqual('test_metric_meta_tag', metric_meta.tags)
        self.assertEqual(metric_meta.properties['a'], metric_meta.properties['a'])
        metric_meta = self.store.update_metric_meta(metric_name=metric_meta.metric_name,
                                                    project_name='test_metric_meta_project_2',
                                                    dataset_name='test_metric_meta_dataset_2')
        metric_meta = self.store.list_dataset_metric_metas(dataset_name=metric_meta.dataset_name)
        self.assertEqual('test_metric_meta_project_2', metric_meta.project_name)
        self.assertEqual('test_metric_meta_dataset_2', metric_meta.dataset_name)
        metric_meta = self.store.list_dataset_metric_metas(project_name=metric_meta.project_name,
                                                           dataset_name=metric_meta.dataset_name)
        self.assertEqual('test_metric_meta_project_2', metric_meta.project_name)
        self.assertEqual('test_metric_meta_dataset_2', metric_meta.dataset_name)
        metric_meta = self.store.register_metric_meta(metric_name='test_metric_meta_2',
                                                      metric_type=MetricType.MODEL,
                                                      metric_desc='test model metric meta',
                                                      project_name='test_metric_meta_project_1',
                                                      model_name='test_metric_meta_model_1',
                                                      start_time=start_time, end_time=end_time,
                                                      uri='/tmp/metric', tags='test_metric_meta_tag',
                                                      properties={'a': 'a'})
        metric_meta = self.store.register_metric_meta(metric_name='test_metric_meta_3',
                                                      metric_type=metric_meta.metric_type,
                                                      metric_desc=metric_meta.metric_desc,
                                                      project_name=metric_meta.project_name,
                                                      model_name=metric_meta.model_name,
                                                      start_time=metric_meta.start_time, end_time=metric_meta.end_time,
                                                      uri=metric_meta.uri, tags=metric_meta.tags,
                                                      properties=metric_meta.properties)
        model_metrics = self.store.list_model_metric_metas(model_name=metric_meta.model_name)
        self.assertEqual(2, len(model_metrics))
        self.assertEqual('test_metric_meta_2', model_metrics[0].metric_name)
        self.assertEqual('test_metric_meta_3', model_metrics[1].metric_name)
        self.assertEqual(2, len(
            self.store.list_model_metric_metas(model_name=metric_meta.model_name,
                                               project_name=metric_meta.project_name)))
        self.store.delete_metric_meta(metric_meta.metric_name)
        model_metric = self.store.list_model_metric_metas(model_name=metric_meta.model_name)
        self.assertEqual('test_metric_meta_2', model_metric.metric_name)

    def test_metric_summary(self):
        metric_timestamp = round(time.time())
        metric_summary = self.store.register_metric_summary(metric_name='test_metric_summary_1', metric_key='auc',
                                                            metric_value='0.6', metric_timestamp=metric_timestamp)
        metric_summary = self.store.get_metric_summary(metric_summary.uuid)
        self.assertEqual(1, metric_summary.uuid)
        self.assertEqual('test_metric_summary_1', metric_summary.metric_name)
        self.assertEqual('auc', metric_summary.metric_key)
        self.assertEqual('0.6', metric_summary.metric_value)
        self.assertEqual(metric_timestamp, metric_summary.metric_timestamp)
        metric_summary = self.store.update_metric_summary(uuid=metric_summary.uuid, metric_value='0.8')
        metric_summary = self.store.get_metric_summary(metric_summary.uuid)
        self.assertEqual('0.8', metric_summary.metric_value)
        metric_summary = self.store.register_metric_summary(metric_name=metric_summary.metric_name,
                                                            metric_key=metric_summary.metric_key,
                                                            metric_value='0.7', metric_timestamp=metric_timestamp + 1,
                                                            model_version='test_metric_summary_model_version_1')
        metric_summary = self.store.register_metric_summary(metric_name=metric_summary.metric_name,
                                                            metric_key='roc',
                                                            metric_value='0.9', metric_timestamp=metric_timestamp + 1,
                                                            model_version='test_metric_summary_model_version_2')
        metric_summaries = self.store.list_metric_summaries(metric_name=metric_summary.metric_name)
        self.assertEqual(3, len(metric_summaries))
        self.assertEqual('auc', metric_summaries[0].metric_key)
        self.assertEqual('0.8', metric_summaries[0].metric_value)
        self.assertEqual('auc', metric_summaries[1].metric_key)
        self.assertEqual('0.7', metric_summaries[1].metric_value)
        self.assertEqual('roc', metric_summaries[2].metric_key)
        self.assertEqual('0.9', metric_summaries[2].metric_value)
        metric_summaries = self.store.list_metric_summaries(metric_key='auc')
        self.assertEqual(2, len(metric_summaries))
        self.assertEqual('0.8', metric_summaries[0].metric_value)
        self.assertEqual('0.7', metric_summaries[1].metric_value)
        metric_summary = self.store.list_metric_summaries(model_version='test_metric_summary_model_version_1')
        self.assertEqual('test_metric_summary_1', metric_summary.metric_name)
        self.assertEqual('auc', metric_summary.metric_key)
        self.assertEqual('0.7', metric_summary.metric_value)
        metric_summary = self.store.list_metric_summaries(model_version='test_metric_summary_model_version_1')
        self.assertEqual('test_metric_summary_1', metric_summary.metric_name)
        self.assertEqual('auc', metric_summary.metric_key)
        self.assertEqual('0.7', metric_summary.metric_value)
        metric_summaries = self.store.list_metric_summaries(metric_name=metric_summary.metric_name,
                                                            start_time=metric_timestamp + 1,
                                                            end_time=metric_summary.metric_timestamp)
        self.assertEqual(2, len(metric_summaries))
        self.assertEqual('auc', metric_summaries[0].metric_key)
        self.assertEqual('0.7', metric_summaries[0].metric_value)
        self.assertEqual('roc', metric_summaries[1].metric_key)
        self.assertEqual('0.9', metric_summaries[1].metric_value)
        metric_summary = self.store.list_metric_summaries(metric_name=metric_summary.metric_name, metric_key='auc',
                                                          model_version='test_metric_summary_model_version_1')
        self.assertEqual('test_metric_summary_1', metric_summary.metric_name)
        self.assertEqual('auc', metric_summary.metric_key)
        self.assertEqual('0.7', metric_summary.metric_value)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from typing import Text, List, Optional, Dict

from ai_flow.common.module_load import import_string
from ai_flow.plugin_interface.job_plugin_interface import JobController, JobHandle
from ai_flow.plugin_interface.scheduler_interface import Scheduler, JobExecutionInfo, WorkflowExecutionInfo, \
    WorkflowInfo
from ai_flow.context.project_context import ProjectContext
from ai_flow.runtime.job_runtime_env import JobRuntimeEnv
from ai_flow.runtime.job_runtime_util import prepare_job_runtime_env
from ai_flow.workflow.status import Status
from ai_flow.workflow.workflow import Workflow, WorkflowPropertyKeys


class SingleJobScheduler(Scheduler):

    def __init__(self, config: Dict):
        super().__init__(config)
        self.workflow: Workflow = None
        self.project_context: ProjectContext = None
        self.job_controller: JobController = None
        self.job_handler: JobHandle = None
        self.job_runtime_env: JobRuntimeEnv = None

    def submit_workflow(self, workflow: Workflow, project_context: ProjectContext) -> WorkflowInfo:
        self.workflow = workflow
        self.project_context = project_context
        return WorkflowInfo(workflow_name=workflow.workflow_name)

    def delete_workflow(self, project_name: Text, workflow_name: Text) -> WorkflowInfo:
        pass

    def pause_workflow_scheduling(self, project_name: Text, workflow_name: Text) -> WorkflowInfo:
        pass

    def resume_workflow_scheduling(self, project_name: Text, workflow_name: Text) -> WorkflowInfo:
        pass

    def start_new_workflow_execution(self, project_name: Text, workflow_name: Text) -> Optional[WorkflowExecutionInfo]:
        pass

    def stop_all_workflow_execution(self, project_name: Text, workflow_name: Text) -> List[WorkflowExecutionInfo]:
        pass

    def stop_workflow_execution(self, execution_id: Text) -> Optional[WorkflowExecutionInfo]:
        pass

    def get_workflow_execution(self, execution_id: Text) -> Optional[WorkflowExecutionInfo]:
        pass

    def list_workflow_executions(self, project_name: Text, workflow_name: Text) -> List[WorkflowExecutionInfo]:
        pass

    def start_job_execution(self, job_name: Text, execution_id: Text) -> JobExecutionInfo:
        job = self.workflow.jobs[job_name]
        plugins = self.workflow.properties.get(WorkflowPropertyKeys.JOB_PLUGINS)
        module, name = plugins.get(job.job_config.job_type)
        class_object = import_string('{}.{}'.format(module, name))
        self.job_controller: JobController = class_object()
        job_execution_info = JobExecutionInfo(job_execution_id='1', job_name=job_name,
                                              workflow_execution
                                              =WorkflowExecutionInfo(workflow_execution_id='1',
                                                                     workflow_info=WorkflowInfo(
                                                                         workflow_name=self.workflow.workflow_name)))
        self.job_runtime_env: JobRuntimeEnv = prepare_job_runtime_env(
            root_working_dir=self.project_context.project_path + '/temp',
            workflow_snapshot_id=self.workflow.workflow_snapshot_id,
            workflow_name=self.workflow.workflow_name,
            job_execution_info=job_execution_info,
            project_context=self.project_context)

        self.job_handler = self.job_controller.submit_job(job=job, job_runtime_env=self.job_runtime_env)
        return job_execution_info

    def stop_job_execution(self, job_name: Text, execution_id: Text) -> JobExecutionInfo:
        self.job_controller.stop_job(self.job_handler, self.job_runtime_env)
        self.job_controller.cleanup_job(self.job_handler, self.job_runtime_env)
        return JobExecutionInfo(job_execution_id='1', job_name=job_name)

    def restart_job_execution(self, job_name: Text, execution_id: Text) -> JobExecutionInfo:
        return JobExecutionInfo(job_execution_id='1', job_name=job_name)

    def get_job_executions(self, job_name: Text, execution_id: Text) -> List[JobExecutionInfo]:
        try:
            result = self.job_controller.get_result(job_handle=self.job_handler)
            print(result)
            self.job_controller.cleanup_job(self.job_handler, self.job_runtime_env)
        except Exception as e:
            return [JobExecutionInfo(job_execution_id='1', job_name=job_name,
                                     status=Status.FAILED, properties={'err': str(e)})]
        return [JobExecutionInfo(job_execution_id='1', job_name=job_name, status=Status.FINISHED)]

    def list_job_executions(self, execution_id: Text) -> List[JobExecutionInfo]:
        pass



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""TaskReschedule tracks rescheduled task instances."""
from sqlalchemy import Column, ForeignKeyConstraint, Index, Integer, String, asc, desc

from airflow.models.base import COLLATION_ARGS, ID_LEN, Base
from airflow.utils.session import provide_session
from airflow.utils.sqlalchemy import UtcDateTime


class TaskReschedule(Base):
    """TaskReschedule tracks rescheduled task instances."""

    __tablename__ = "task_reschedule"

    id = Column(Integer, primary_key=True)
    task_id = Column(String(ID_LEN, **COLLATION_ARGS), nullable=False)
    dag_id = Column(String(ID_LEN, **COLLATION_ARGS), nullable=False)
    execution_date = Column(UtcDateTime, nullable=False)
    try_number = Column(Integer, nullable=False)
    start_date = Column(UtcDateTime, nullable=False)
    end_date = Column(UtcDateTime, nullable=False)
    duration = Column(Integer, nullable=False)
    reschedule_date = Column(UtcDateTime, nullable=False)

    __table_args__ = (
        Index('idx_task_reschedule_dag_task_date', dag_id, task_id, execution_date, unique=False),
        ForeignKeyConstraint(
            [task_id, dag_id, execution_date],
            ['task_instance.task_id', 'task_instance.dag_id', 'task_instance.execution_date'],
            name='task_reschedule_dag_task_date_fkey',
            ondelete='CASCADE',
        ),
    )

    def __init__(self, task, execution_date, try_number, start_date, end_date, reschedule_date):
        self.dag_id = task.dag_id
        self.task_id = task.task_id
        self.execution_date = execution_date
        self.try_number = try_number
        self.start_date = start_date
        self.end_date = end_date
        self.reschedule_date = reschedule_date
        self.duration = (self.end_date - self.start_date).total_seconds()

    @staticmethod
    @provide_session
    def query_for_task_instance(task_instance, descending=False, session=None):
        """
        Returns query for task reschedules for a given the task instance.

        :param session: the database session object
        :type session: sqlalchemy.orm.session.Session
        :param task_instance: the task instance to find task reschedules for
        :type task_instance: airflow.models.TaskInstance
        :param descending: If True then records are returned in descending order
        :type descending: bool
        """
        TR = TaskReschedule
        qry = session.query(TR).filter(
            TR.dag_id == task_instance.dag_id,
            TR.task_id == task_instance.task_id,
            TR.execution_date == task_instance.execution_date,
            TR.try_number == task_instance.try_number,
        )
        if descending:
            return qry.order_by(desc(TR.id))
        else:
            return qry.order_by(asc(TR.id))

    @staticmethod
    @provide_session
    def find_for_task_instance(task_instance, session=None):
        """
        Returns all task reschedules for the task instance and try number,
        in ascending order.

        :param session: the database session object
        :type session: sqlalchemy.orm.session.Session
        :param task_instance: the task instance to find task reschedules for
        :type task_instance: airflow.models.TaskInstance
        """
        return TaskReschedule.query_for_task_instance(task_instance, session=session).all()



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""This module is deprecated. Please use `airflow.providers.google.cloud.operators.dlp`."""

import warnings

# pylint: disable=unused-import
from airflow.providers.google.cloud.operators.dlp import (  # noqa
    CloudDLPCancelDLPJobOperator,
    CloudDLPCreateDeidentifyTemplateOperator,
    CloudDLPCreateDLPJobOperator,
    CloudDLPCreateInspectTemplateOperator,
    CloudDLPCreateJobTriggerOperator,
    CloudDLPCreateStoredInfoTypeOperator,
    CloudDLPDeidentifyContentOperator,
    CloudDLPDeleteDeidentifyTemplateOperator,
    CloudDLPDeleteDLPJobOperator,
    CloudDLPDeleteInspectTemplateOperator,
    CloudDLPDeleteJobTriggerOperator,
    CloudDLPDeleteStoredInfoTypeOperator,
    CloudDLPGetDeidentifyTemplateOperator,
    CloudDLPGetDLPJobOperator,
    CloudDLPGetDLPJobTriggerOperator,
    CloudDLPGetInspectTemplateOperator,
    CloudDLPGetStoredInfoTypeOperator,
    CloudDLPInspectContentOperator,
    CloudDLPListDeidentifyTemplatesOperator,
    CloudDLPListDLPJobsOperator,
    CloudDLPListInfoTypesOperator,
    CloudDLPListInspectTemplatesOperator,
    CloudDLPListJobTriggersOperator,
    CloudDLPListStoredInfoTypesOperator,
    CloudDLPRedactImageOperator,
    CloudDLPReidentifyContentOperator,
    CloudDLPUpdateDeidentifyTemplateOperator,
    CloudDLPUpdateInspectTemplateOperator,
    CloudDLPUpdateJobTriggerOperator,
    CloudDLPUpdateStoredInfoTypeOperator,
)

warnings.warn(
    "This module is deprecated. Please use `airflow.providers.google.cloud.operators.dlp`.",
    DeprecationWarning,
    stacklevel=2,
)


class CloudDLPDeleteDlpJobOperator(CloudDLPDeleteDLPJobOperator):
    """
    This class is deprecated.
    Please use `airflow.providers.google.cloud.operators.dlp.CloudDLPDeleteDLPJobOperator`.
    """

    def __init__(self, *args, **kwargs):
        warnings.warn(
            """This class is deprecated.
            Please use `airflow.providers.google.cloud.operators.dlp.CloudDLPDeleteDLPJobOperator`.""",
            DeprecationWarning,
            stacklevel=3,
        )
        super().__init__(*args, **kwargs)


class CloudDLPGetDlpJobOperator(CloudDLPGetDLPJobOperator):
    """
    This class is deprecated.
    Please use `airflow.providers.google.cloud.operators.dlp.CloudDLPGetDLPJobOperator`.
    """

    def __init__(self, *args, **kwargs):
        warnings.warn(
            """This class is deprecated.
            Please use `airflow.providers.google.cloud.operators.dlp.CloudDLPGetDLPJobOperator`.""",
            DeprecationWarning,
            stacklevel=3,
        )
        super().__init__(*args, **kwargs)


class CloudDLPGetJobTripperOperator(CloudDLPGetDLPJobTriggerOperator):
    """
    This class is deprecated.
    Please use `airflow.providers.google.cloud.operators.dlp.CloudDLPGetDLPJobTriggerOperator`.
    """

    def __init__(self, *args, **kwargs):
        warnings.warn(
            """This class is deprecated.
            Please use `airflow.providers.google.cloud.operators.dlp.CloudDLPGetDLPJobTriggerOperator`.""",
            DeprecationWarning,
            stacklevel=3,
        )
        super().__init__(*args, **kwargs)


class CloudDLPListDlpJobsOperator(CloudDLPListDLPJobsOperator):
    """
    This class is deprecated.
    Please use `airflow.providers.google.cloud.operators.dlp.CloudDLPListDLPJobsOperator`.
    """

    def __init__(self, *args, **kwargs):
        warnings.warn(
            """This class is deprecated.
            Please use `airflow.providers.google.cloud.operators.dlp.CloudDLPListDLPJobsOperator`.""",
            DeprecationWarning,
            stacklevel=3,
        )
        super().__init__(*args, **kwargs)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
import tempfile
from typing import Text, Dict, Any
from pathlib import Path
import oss2
from ai_flow.plugin_interface.blob_manager_interface import BlobManager
from ai_flow.util.file_util.zip_file_util import make_dir_zipfile
from ai_flow_plugins.blob_manager_plugins.blob_manager_utils import extract_project_zip_file


class OssBlobManager(BlobManager):
    """
    OssBlobManager is an implementation of BlobManager based on the oss file system
    OssBlobManager contains configuration items:
    1. access_key_id: The oss access key.
    2. access_key_secret: The oss access secret.
    3. bucket: The oss bucket name.
    4. local_repository: It represents the root path of the downloaded project package.
    """
    def __init__(self, config: Dict[str, Any]):
        super().__init__(config)
        ack_id = config.get('access_key_id', None)
        ack_secret = config.get('access_key_secret', None)
        endpoint = config.get('endpoint', None)
        bucket_name = config.get('bucket', None)
        auth = oss2.Auth(ack_id, ack_secret)
        self.bucket = oss2.Bucket(auth, endpoint, bucket_name)
        self.repo_name = config.get('repo_name', '')
        self._local_repo = config.get('local_repository', None)

    def upload_project(self, workflow_snapshot_id: Text, project_path: Text) -> Text:
        """
        upload a given project to blob server for remote execution.

        :param workflow_snapshot_id: It is the unique identifier for each workflow generation.
        :param project_path: The path of this project.
        :return The uri of the uploaded project file in blob server.
        """
        with tempfile.TemporaryDirectory() as temp_dir:
            zip_file_name = 'workflow_{}_project.zip'.format(workflow_snapshot_id)
            temp_dir_path = Path(temp_dir)
            zip_file_path = temp_dir_path / zip_file_name
            make_dir_zipfile(project_path, zip_file_path)
            object_key = self.repo_name + '/' + zip_file_name
            self.bucket.put_object_from_file(key=object_key, filename=str(zip_file_path))
        return object_key

    def download_project(self, workflow_snapshot_id, remote_path: Text, local_path: Text = None) -> Text:
        """
        download the needed resource from remote blob server to local process for remote execution.

        :param workflow_snapshot_id: It is the unique identifier for each workflow generation.
        :param remote_path: The project package uri.
        :param local_path: Download file root path.
        :return The local project path.
        """
        local_zip_file_name = 'workflow_{}_project'.format(workflow_snapshot_id)
        oss_object_key = remote_path
        if local_path is not None:
            repo_path = Path(local_path)
        elif self._local_repo is not None:
            repo_path = Path(self._local_repo)
        else:
            repo_path = Path(tempfile.gettempdir())
        local_zip_file_path = str(repo_path / local_zip_file_name) + '.zip'
        extract_path = str(repo_path / local_zip_file_name)
        self.bucket.get_object_to_file(oss_object_key, filename=local_zip_file_path)
        return extract_project_zip_file(workflow_snapshot_id=workflow_snapshot_id,
                                        local_root_path=repo_path,
                                        zip_file_path=local_zip_file_path,
                                        extract_project_path=extract_path)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""Connection sub-commands"""
import io
import json
import os
import sys
from typing import Any, Dict, List
from urllib.parse import urlparse, urlunparse

import yaml
from sqlalchemy.orm import exc

from airflow.cli.simple_table import AirflowConsole
from airflow.exceptions import AirflowNotFoundException
from airflow.hooks.base import BaseHook
from airflow.models import Connection
from airflow.utils import cli as cli_utils
from airflow.utils.cli import suppress_logs_and_warning
from airflow.utils.session import create_session


def _connection_mapper(conn: Connection) -> Dict[str, Any]:
    return {
        'id': conn.id,
        'conn_id': conn.conn_id,
        'conn_type': conn.conn_type,
        'description': conn.description,
        'host': conn.host,
        'schema': conn.schema,
        'login': conn.login,
        'password': conn.password,
        'port': conn.port,
        'is_encrypted': conn.is_encrypted,
        'is_extra_encrypted': conn.is_encrypted,
        'extra_dejson': conn.extra_dejson,
        'get_uri': conn.get_uri(),
    }


@suppress_logs_and_warning()
def connections_get(args):
    """Get a connection."""
    try:
        conn = BaseHook.get_connection(args.conn_id)
    except AirflowNotFoundException:
        raise SystemExit("Connection not found.")
    AirflowConsole().print_as(
        data=[conn],
        output=args.output,
        mapper=_connection_mapper,
    )


@suppress_logs_and_warning()
def connections_list(args):
    """Lists all connections at the command line"""
    with create_session() as session:
        query = session.query(Connection)
        if args.conn_id:
            query = query.filter(Connection.conn_id == args.conn_id)
        conns = query.all()

        AirflowConsole().print_as(
            data=conns,
            output=args.output,
            mapper=_connection_mapper,
        )


def _format_connections(conns: List[Connection], fmt: str) -> str:
    if fmt == '.env':
        connections_env = ""
        for conn in conns:
            connections_env += f"{conn.conn_id}={conn.get_uri()}\n"
        return connections_env

    connections_dict = {}
    for conn in conns:
        connections_dict[conn.conn_id] = {
            'conn_type': conn.conn_type,
            'description': conn.description,
            'host': conn.host,
            'login': conn.login,
            'password': conn.password,
            'schema': conn.schema,
            'port': conn.port,
            'extra': conn.extra,
        }

    if fmt == '.yaml':
        return yaml.dump(connections_dict)

    if fmt == '.json':
        return json.dumps(connections_dict, indent=2)

    return json.dumps(connections_dict)


def _is_stdout(fileio: io.TextIOWrapper) -> bool:
    return fileio.name == '<stdout>'


def _valid_uri(uri: str) -> bool:
    """Check if a URI is valid, by checking if both scheme and netloc are available"""
    uri_parts = urlparse(uri)
    return uri_parts.scheme != '' and uri_parts.netloc != ''


def connections_export(args):
    """Exports all connections to a file"""
    allowed_formats = ['.yaml', '.json', '.env']
    provided_format = None if args.format is None else f".{args.format.lower()}"
    default_format = provided_format or '.json'

    with create_session() as session:
        if _is_stdout(args.file):
            filetype = default_format
        elif provided_format is not None:
            filetype = provided_format
        else:
            _, filetype = os.path.splitext(args.file.name)
            filetype = filetype.lower()
            if filetype not in allowed_formats:
                raise SystemExit(
                    f"Unsupported file format. The file must have "
                    f"the extension {', '.join(allowed_formats)}."
                )

        connections = session.query(Connection).order_by(Connection.conn_id).all()
        msg = _format_connections(connections, filetype)
        args.file.write(msg)

        if _is_stdout(args.file):
            print("Connections successfully exported.", file=sys.stderr)
        else:
            print(f"Connections successfully exported to {args.file.name}.")


alternative_conn_specs = ['conn_type', 'conn_host', 'conn_login', 'conn_password', 'conn_schema', 'conn_port']


@cli_utils.action_logging
def connections_add(args):
    """Adds new connection"""
    # Check that the conn_id and conn_uri args were passed to the command:
    missing_args = []
    invalid_args = []
    if args.conn_uri:
        if not _valid_uri(args.conn_uri):
            raise SystemExit(f'The URI provided to --conn-uri is invalid: {args.conn_uri}')
        for arg in alternative_conn_specs:
            if getattr(args, arg) is not None:
                invalid_args.append(arg)
    elif not args.conn_type:
        missing_args.append('conn-uri or conn-type')
    if missing_args:
        raise SystemExit(f'The following args are required to add a connection: {missing_args!r}')
    if invalid_args:
        raise SystemExit(
            f'The following args are not compatible with the '
            f'add flag and --conn-uri flag: {invalid_args!r}'
        )

    if args.conn_uri:
        new_conn = Connection(conn_id=args.conn_id, description=args.conn_description, uri=args.conn_uri)
    else:
        new_conn = Connection(
            conn_id=args.conn_id,
            conn_type=args.conn_type,
            description=args.conn_description,
            host=args.conn_host,
            login=args.conn_login,
            password=args.conn_password,
            schema=args.conn_schema,
            port=args.conn_port,
        )
    if args.conn_extra is not None:
        new_conn.set_extra(args.conn_extra)

    with create_session() as session:
        if not session.query(Connection).filter(Connection.conn_id == new_conn.conn_id).first():
            session.add(new_conn)
            msg = 'Successfully added `conn_id`={conn_id} : {uri}'
            msg = msg.format(
                conn_id=new_conn.conn_id,
                uri=args.conn_uri
                or urlunparse(
                    (
                        args.conn_type,
                        '{login}:{password}@{host}:{port}'.format(
                            login=args.conn_login or '',
                            password='******' if args.conn_password else '',
                            host=args.conn_host or '',
                            port=args.conn_port or '',
                        ),
                        args.conn_schema or '',
                        '',
                        '',
                        '',
                    )
                ),
            )
            print(msg)
        else:
            msg = f'A connection with `conn_id`={new_conn.conn_id} already exists.'
            raise SystemExit(msg)


@cli_utils.action_logging
def connections_delete(args):
    """Deletes connection from DB"""
    with create_session() as session:
        try:
            to_delete = session.query(Connection).filter(Connection.conn_id == args.conn_id).one()
        except exc.NoResultFound:
            raise SystemExit(f'Did not find a connection with `conn_id`={args.conn_id}')
        except exc.MultipleResultsFound:
            raise SystemExit(f'Found more than one connection with `conn_id`={args.conn_id}')
        else:
            session.delete(to_delete)
            print(f"Successfully deleted connection with `conn_id`={to_delete.conn_id}")



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""Scheduler command"""
import signal

import daemon
from airflow.contrib.jobs.event_based_scheduler_job import EventBasedSchedulerJob
from airflow.contrib.jobs.scheduler_factory import SchedulerFactory
from daemon.pidfile import TimeoutPIDLockFile

from airflow import settings
from airflow.jobs.scheduler_job import SchedulerJob
from airflow.utils import cli as cli_utils
from airflow.utils.cli import process_subdir, setup_locations, setup_logging, sigint_handler, sigquit_handler


@cli_utils.action_logging
def scheduler(args):
    """Starts Airflow Scheduler"""
    print(settings.HEADER)
    job = SchedulerJob(
        subdir=process_subdir(args.subdir),
        num_runs=args.num_runs,
        do_pickle=args.do_pickle,
    )
    scheduler_name = SchedulerFactory.get_scheduler_name()
    if scheduler_name == SchedulerFactory.DEFAULT_SCHEDULER:
        pass
    elif scheduler_name == SchedulerFactory.EVENT_BASED_SCHEDULER:
        job = EventBasedSchedulerJob(
            dag_directory=process_subdir(args.subdir),
            server_uri=args.server_uri)
    else:
        scheduler_class = SchedulerFactory.get_default_scheduler()
        job = scheduler_class()

    if args.daemon:
        pid, stdout, stderr, log_file = setup_locations(
            "scheduler", args.pid, args.stdout, args.stderr, args.log_file
        )
        handle = setup_logging(log_file)
        stdout = open(stdout, 'w+')
        stderr = open(stderr, 'w+')

        ctx = daemon.DaemonContext(
            pidfile=TimeoutPIDLockFile(pid, -1),
            files_preserve=[handle],
            stdout=stdout,
            stderr=stderr,
        )
        with ctx:
            job.run()

        stdout.close()
        stderr.close()
    else:
        signal.signal(signal.SIGINT, sigint_handler)
        signal.signal(signal.SIGTERM, sigint_handler)
        signal.signal(signal.SIGQUIT, sigquit_handler)
        job.run()



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""
Example usage of the TriggerDagRunOperator. This example holds 2 DAGs:
1. 1st DAG (example_trigger_controller_dag) holds a TriggerDagRunOperator, which will trigger the 2nd DAG
2. 2nd DAG (example_trigger_target_dag) which will be triggered by the TriggerDagRunOperator in the 1st DAG
"""

from airflow import DAG
from airflow.operators.bash import BashOperator
from airflow.operators.python import PythonOperator
from airflow.utils.dates import days_ago

dag = DAG(
    dag_id="example_trigger_target_dag",
    default_args={"owner": "airflow"},
    start_date=days_ago(2),
    schedule_interval=None,
    tags=['example'],
)


def run_this_func(**context):
    """
    Print the payload "message" passed to the DagRun conf attribute.

    :param context: The execution context
    :type context: dict
    """
    print("Remotely received value of {} for key=message".format(context["dag_run"].conf["message"]))


run_this = PythonOperator(task_id="run_this", python_callable=run_this_func, dag=dag)

bash_task = BashOperator(
    task_id="bash_task",
    bash_command='echo "Here is the message: $message"',
    env={'message': '{{ dag_run.conf["message"] if dag_run else "" }}'},
    dag=dag,
)
